'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 1.0.2

version: 1.0.2

full_version: 1.0.2

git_revision: 949600e6def935b323c822b221c3258420aee441

release: True

'''
short_version: str = '1.0.2'
version: str = '1.0.2'
full_version: str = '1.0.2'
git_revision: str = '949600e6def935b323c822b221c3258420aee441'
release: bool = True
if not release:
    version = full_version



"""
The `MultiPnL` class allows for the visualization and analysis of PnLs across multiple return categories.
It also provides functionality to calculate a weighted aggregate PnL based on user-defined weights
for each PnL.
"""

from typing import Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import update_df, _map_to_business_day_frequency
from macrosynergy.pnl import NaivePnL
from macrosynergy.management.types import QuantamentalDataFrame


class MultiPnL:
    """
    Manages multiple `NaivePnL` instances, enabling combined PnL analysis and visualization.
    """

    def __init__(self):
        self.pnls_df = QuantamentalDataFrame(
            pd.DataFrame(columns=["real_date", "xcat", "value", "cid"])
        )
        self.single_return_pnls: Dict[str, NaivePnL] = {}
        self.composite_pnl_xcats = []
        self.xcat_to_ret = {}

    def add_pnl(self, pnl: NaivePnL, pnl_xcats: List[str]):
        """
        Add PnL(s) from a NaivePnL object. PnL categories will be ingested into
        the MultiPnL object as 'pnl_xcat/return'.

        Parameters
        ----------
        pnl : NaivePnL
            NaivePnL object.
        pnl_xcats : List[str]
            List of PnLs to add from the NaivePnL object.
        """

        self._validate_pnl(pnl, pnl_xcats)

        pnl_df = pnl.pnl_df(pnl_xcats)
        pnl_df = QuantamentalDataFrame(pnl_df).rename_xcats(postfix=f"/{pnl.ret}")
        self.pnls_df = update_df(self.pnls_df, pnl_df)
        for xcat in pnl_df.xcat.unique():
            self.single_return_pnls[xcat] = pnl

        for xcat in pnl_xcats:
            if xcat not in self.xcat_to_ret:
                self.xcat_to_ret[xcat] = {pnl.ret}
            else:
                self.xcat_to_ret[xcat].add(pnl.ret)
        pass

    def combine_pnls(
        self,
        pnl_xcats: List[str],
        composite_pnl_xcat: str,
        weights: Optional[Dict[str, float]] = None,
    ):
        """
        Combine PnLs with optional weighting.

        Parameters
        ----------
        pnl_xcats : List[str]
            List of PnLs to combine. Must be in the format 'xcat/return' and added using
            `add_pnl()`.
        composite_pnl_xcat : str
            xcat for the combined PnL.
        weights : Optional[Dict[str, float]]
            Weights for each PnL, by default None. Must be in the format {'xcat':
            weight} or {'xcat/return': weight}.
        """

        self._check_pnls_added(min_pnls=2)
        for i, pnl_xcat in enumerate(pnl_xcats):
            pnl_xcats[i] = self._infer_return_by_xcat(pnl_xcat)

        # Default weights
        if weights is None:
            weights = {pnl_name: 1 for pnl_name in pnl_xcats}
        else:
            weights = {self._infer_return_by_xcat(k): v for k, v in weights.items()}
        weights = self._normalize_weights(weights)

        multiasset_df = []
        for pnl_xcat in pnl_xcats:
            single_asset_df = self.pnls_df[self.pnls_df["xcat"] == pnl_xcat].assign(
                asset=pnl_xcat
            )
            multiasset_df.append(single_asset_df)

        multiasset_df = QuantamentalDataFrame.from_qdf_list(multiasset_df)

        raw_pnls = multiasset_df.set_index(["real_date", "xcat"])["value"].unstack()

        # Default weights for each strategy
        start_weights = pd.DataFrame(
            {asset_name: weights[asset_name] for asset_name in raw_pnls.columns},
            index=raw_pnls.index,
        )

        # Daily change in portfolio weights due to previous returns since the last rebalancing
        mfreq = _map_to_business_day_frequency("M")
        weights_change = (
            (1 + raw_pnls / 100).groupby(pd.Grouper(freq=mfreq)).cumprod()
        )  # in decimals, not percentage, gross amount
        weights_change = (
            weights_change.groupby(pd.Grouper(freq=mfreq))
            .shift(periods=1)
            .fillna(value=1)
        )

        # Dynamic weights
        final_weights = start_weights * weights_change
        final_weights = final_weights.div(final_weights.sum(axis=1), axis=0)

        # final calculation
        multiasset_rets = (final_weights * raw_pnls).sum(axis=1)

        multiasset_rets.name = composite_pnl_xcat

        multi_asset_pnl = QuantamentalDataFrame.from_long_df(
            multiasset_rets.reset_index().melt(
                id_vars=["real_date"], var_name="xcat", value_name="value"
            ),
            cid="ALL",
        )
        multi_asset_pnl = multi_asset_pnl.sort_values(by=["xcat", "real_date"])

        self.pnls_df = update_df(self.pnls_df, multi_asset_pnl).sort_values(
            by=["xcat", "real_date"]
        )
        self.composite_pnl_xcats.append(composite_pnl_xcat)

    def plot_pnls(
        self,
        pnl_xcats: List[str] = None,
        title: str = None,
        title_fontsize: int = 14,
        xcat_labels: Union[List[str], dict] = None,
    ):
        """
        Creates a plot of PnLs from added NaivePnL objects and/or
        combined PnLs created with `combine_pnls()`.

        Parameters
        ----------
        pnl_xcats : List[str]
            List of PnLs to plot. If None, all PnLs are plotted. Must be in the format
            'xcat', or 'xcat/return_xcat'.
        title : str
            Title of the plot.
        """

        self._check_pnls_added()

        if pnl_xcats is None:
            pnl_df = self.pnls_df
            pnl_xcats = self.pnl_xcats
        else:
            for i, pnl_xcat in enumerate(pnl_xcats):
                pnl_xcats[i] = self._infer_return_by_xcat(pnl_xcat)
            pnl_df = self.pnls_df[self.pnls_df["xcat"].isin(pnl_xcats)].copy()

        if pnl_df["cid"].dtype.name == "category":
            pnl_df.loc[:, "cid"] = pnl_df["cid"].cat.remove_unused_categories()
        if pnl_df["xcat"].dtype.name == "category":
            pnl_df.loc[:, "xcat"] = pnl_df["xcat"].cat.remove_unused_categories()

        if xcat_labels is not None:
            xcat_labels = self._check_xcat_labels(pnl_xcats, xcat_labels)
            pnl_df["xcat"] = pnl_df["xcat"].map(xcat_labels)

        pnl_df.loc[:, "cumulative pnl"] = pnl_df.groupby("xcat")["value"].cumsum()

        sns.lineplot(data=pnl_df, x="real_date", y="cumulative pnl", hue=("xcat"))
        plt.title(title, fontsize=title_fontsize)
        plt.xlabel(None)
        plt.ylabel("% risk capital, no compounding")
        plt.legend(title="PnL Category(s)")
        plt.show()
        pnl_df.drop(columns="cumulative pnl", inplace=True)

    def evaluate_pnls(self, pnl_xcats: List[str] = None) -> pd.DataFrame:
        """
        Returns a DataFrame containing the following evaluation metrics for specified PnLs:

        - Return %
        - St. Dev. %
        - Sharpe Ratio
        - Sortino Ratio
        - Max 21-Day Draw %
        - Max 6-Month Draw %
        - Peak to Trough Draw %
        - Top 5% Monthly PnL Share
        - Traded Months
        - Correlation with benchmarks (if available)

        .. note::
            The evaluation metrics are calculated in a manner similar to NaivePnL's `evaluate_pnls()`.

            Benchmark correlation is not currently supported for combined PnLs.

        Parameters
        ----------
        pnl_xcats : List[str]
            List of PnLs to evaluate. If None, all PnLs are evaluated. Must be in the
            format 'xcat', or 'xcat/return_xcat'.

        Returns
        -------
        ~pandas.DataFrame
            DataFrame containing evaluation metrics for the specified PnLs.
        """

        self._check_pnls_added()
        if pnl_xcats is None:
            pnl_xcats = self.pnl_xcats
        else:
            for i, pnl_xcat in enumerate(pnl_xcats):
                pnl_xcats[i] = self._infer_return_by_xcat(pnl_xcat)
        pnl_evals = []
        for pnl_xcat in pnl_xcats:
            if pnl_xcat in self.composite_pnl_xcats:
                eval_df = self._evaluate_composite_pnl(pnl_xcat)
            else:
                pnl = self.single_return_pnls[pnl_xcat]
                eval_df = pnl.evaluate_pnls([pnl_xcat.split("/")[0]])
                eval_df.columns = [pnl_xcat]
            pnl_evals.append(eval_df)

        return pd.concat(pnl_evals, axis=1, ignore_index=False, sort=False)

    def _evaluate_composite_pnl(self, pnl_xcat: str) -> pd.DataFrame:
        """
        Evaluate the combined PnLs in a manner similar to NaivePnL's `evaluate_pnls()`.
        """
        stats = [
            "Return %",
            "St. Dev. %",
            "Sharpe Ratio",
            "Sortino Ratio",
            "Max 21-Day Draw %",
            "Max 6-Month Draw %",
            "Peak to Trough Draw %",
            "Top 5% Monthly PnL Share",
            "Traded Months",
        ]
        pnl_df = self.get_pnls([pnl_xcat])
        dfw = pnl_df.pivot(index="real_date", columns="xcat", values="value")
        df = pd.DataFrame(columns=dfw.columns, index=stats)

        df.iloc[0, :] = dfw.mean(axis=0) * 261
        df.iloc[1, :] = dfw.std(axis=0) * np.sqrt(261)
        df.iloc[2, :] = df.iloc[0, :] / df.iloc[1, :]
        dsd = dfw.apply(lambda x: np.sqrt(np.sum(x[x < 0] ** 2) / len(x))) * np.sqrt(
            261
        )
        df.iloc[3, :] = df.iloc[0, :] / dsd
        df.iloc[4, :] = dfw.rolling(21).sum().min()
        df.iloc[5, :] = dfw.rolling(6 * 21).sum().min()

        cum_pnl = dfw.cumsum()
        high_watermark = cum_pnl.cummax()
        drawdown = high_watermark - cum_pnl

        df.iloc[6, :] = -drawdown.max()

        monthly_pnl = dfw.resample("M").sum()
        total_pnl = monthly_pnl.sum(axis=0)
        top_5_percent_cutoff = int(np.ceil(len(monthly_pnl) * 0.05))
        top_months = pd.DataFrame(columns=monthly_pnl.columns)
        for column in monthly_pnl.columns:
            top_months[column] = (
                monthly_pnl[column]
                .nlargest(top_5_percent_cutoff)
                .reset_index(drop=True)
            )
        df.iloc[7, :] = top_months.sum() / total_pnl
        df.iloc[8, :] = dfw.resample("M").sum().count()
        return df

    def get_pnls(self, pnl_xcats: List[str] = None) -> pd.DataFrame:
        """
        Returns a DataFrame containing specified PnLs.

        Parameters
        ----------
        pnl_xcats : List[str]
            List of PnLs to return. If None, all PnLs are returned. Must be in the
            format 'xcat', or 'xcat/return_xcat'.

        Returns
        -------
        ~pandas.DataFrame
            DataFrame containing the specified PnLs.
        """

        if self.pnls_df is None:
            raise ValueError("The PnLs have been added. Use add_pnl() first.")

        if pnl_xcats is None:
            return QuantamentalDataFrame(self.pnls_df, categorical=False)

        else:
            for i, pnl_xcat in enumerate(pnl_xcats):
                pnl_xcats[i] = self._infer_return_by_xcat(pnl_xcat)

            return (
                QuantamentalDataFrame(self.pnls_df, _initialized_as_categorical=False)
                .reduce_df(xcats=pnl_xcats)
                .to_original_dtypes()
            )

    def _normalize_weights(self, weights: dict) -> dict:
        """
        Normalize the weights to sum up to 1.
        """
        weights_sum = sum(weights.values())
        return {k: v / weights_sum for k, v in weights.items()}

    def _validate_pnl(self, pnl: NaivePnL, pnl_xcats: List[str]):
        """
        Validate the PnL and PnL categories.
        """
        if not isinstance(pnl, NaivePnL):
            raise ValueError("The pnl must be a NaivePnL object.")
        if not set(pnl_xcats).issubset(pnl.pnl_names):
            raise ValueError("The pnl_xcats must be in the NaivePnL object.")
        if not all(isinstance(x, str) for x in pnl_xcats):
            raise ValueError("All elements in the list must be strings.")
        return True

    @property
    def pnl_xcats(self):
        """
        List of all unique PnL categories stored in the MultiPnL object.

        Returns
        -------
        List[str]
            A list of unique category names from the "xcat" column in `pnls_df`.
        """
        return self.pnls_df["xcat"].unique().tolist()

    @property
    def return_xcats(self):
        """
        List of all unique return categories associated with PnLs.

        Returns
        -------
        List[str]
            A list of unique return category names derived from the `xcat_to_ret` mapping.
        """
        return list(set(self.xcat_to_ret.values()))

    def _check_pnls_added(self, min_pnls: int = 1):
        """
        Check if at least `min_pnls` PnLs have been added.
        """
        if len(self.pnl_xcats) < min_pnls:
            raise ValueError(
                f"At least {min_pnls} PnL must be added with add_pnl() first."
            )

    def _infer_return_by_xcat(self, pnl_xcat):
        """
        Infer the return category from the xcat if not provided. Throws an error if there
        are multiple return categories for the xcat.
        """

        if pnl_xcat in self.composite_pnl_xcats:
            return pnl_xcat

        if "/" not in pnl_xcat:
            if pnl_xcat not in self.xcat_to_ret:
                raise ValueError(f"{pnl_xcat} has not been added with add_pnl() yet.")
            if len(self.xcat_to_ret[pnl_xcat]) > 1:
                raise ValueError(
                    f"{pnl_xcat} corresponds to multiple return categories: {self.xcat_to_ret[pnl_xcat]}. "
                    "Must append return to xcat in the format 'xcat/return'."
                )
            else:
                return f"{pnl_xcat}/{list(self.xcat_to_ret[pnl_xcat])[0]}"
        else:
            if pnl_xcat not in self.pnl_xcats:
                raise ValueError(f"{pnl_xcat} has not been added with add_pnl() yet.")
            else:
                return pnl_xcat

    def _check_xcat_labels(self, pnl_xcats, xcat_labels):
        if isinstance(xcat_labels, dict):
            xcat_labels = {
                self._infer_return_by_xcat(k): v for k, v in xcat_labels.items()
            }
        elif isinstance(xcat_labels, list):
            if len(pnl_xcats) != len(xcat_labels):
                raise ValueError(
                    "If using a list, the number of labels must match the number of PnLs."
                )
            xcat_labels = dict(zip(pnl_xcats, xcat_labels))
        else:
            raise ValueError("xcat_labels must be a list or a dictionary.")
        return xcat_labels


if __name__ == "__main__":
    np.random.seed(0)

    cids = ["AUD", "CAD", "GBP", "NZD", "USD", "EUR"]
    xcats = ["EQXR_NSA", "FXXR", "GROWTH", "INFL", "DUXR"]

    cols_1 = ["earliest", "latest", "mean_add", "sd_mult"]
    df_cids = pd.DataFrame(index=cids, columns=cols_1)
    df_cids.loc["AUD", :] = ["2008-01-03", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD", :] = ["2010-01-03", "2020-11-30", 0, 1]
    df_cids.loc["GBP", :] = ["2012-01-03", "2020-11-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-03", "2020-09-30", -0.1, 2]
    df_cids.loc["USD"] = ["2015-01-03", "2020-12-31", 0.2, 2]
    df_cids.loc["EUR"] = ["2008-01-03", "2020-12-31", 0.1, 2]

    cols_2 = cols_1 + ["ar_coef", "back_coef"]

    df_xcats = pd.DataFrame(index=xcats, columns=cols_2)
    df_xcats.loc["EQXR_NSA"] = ["2000-01-03", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["FXXR"] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-03", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]
    df_xcats.loc["DUXR"] = ["2000-01-01", "2020-12-31", 0.1, 0.5, 0, 0.1]

    black = {"AUD": ["2006-01-01", "2015-12-31"], "GBP": ["2022-01-01", "2100-01-01"]}
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Instantiate a new instance to test the long-only functionality.
    # Benchmarks are used to calculate correlation against PnL series.
    pnl_eq = NaivePnL(
        dfd,
        ret="EQXR_NSA",
        sigs=["GROWTH"],
        cids=cids,
        start="2000-01-01",
        blacklist=black,
        # bms=["EUR_EQXR_NSA", "USD_EQXR_NSA"],
    )

    pnl_eq.make_pnl(
        sig="GROWTH",
        sig_op="zn_score_pan",
        sig_neg=False,
        sig_add=0.5,
        rebal_freq="monthly",
        vol_scale=5,
        rebal_slip=1,
        min_obs=250,
        thresh=2,
        pnl_name="PNL_EQ",
    )

    pnl_eq.make_long_pnl(vol_scale=10, label="LONG")

    pnl_fx = NaivePnL(
        dfd,
        ret="FXXR",
        sigs=["INFL"],
        cids=cids,
        start="2000-01-01",
        blacklist=black,
        bms=["EUR_EQXR_NSA", "USD_EQXR_NSA"],
    )

    pnl_fx.make_pnl(
        sig="INFL",
        sig_op="zn_score_pan",
        sig_neg=True,
        sig_add=0.5,
        rebal_freq="monthly",
        vol_scale=5,
        rebal_slip=1,
        min_obs=250,
        thresh=2,
        pnl_name="PNL_FX",
    )

    pnl_fx.make_long_pnl(vol_scale=10, label="LONG")

    # pnl_fx.make_long_pnl(vol_scale=10, label="Long")
    print(pnl_eq.pnl_names)

    mapnl = MultiPnL()

    mapnl.add_pnl(pnl_fx, ["PNL_FX", "LONG"])
    mapnl.add_pnl(pnl_eq, ["PNL_EQ", "LONG"])

    mapnl.combine_pnls(
        # ["PNL_FX", "PNL_EQ"],
        ["PNL_EQ", "PNL_FX"],
        # weights={"PNL_FX": 1, "LONG": 1},
        composite_pnl_xcat="EQ_FX_LONG",
    )
    mapnl.evaluate_pnls(["EQ_FX_LONG"])
    mapnl.plot_pnls(["PNL_FX", "PNL_EQ"], xcat_labels=["z", "FX"], title="PnLs")
    # print(mapnl.get_pnls(["PNL_FX"]))
    # print(mapnl.evaluate_pnls(["PNL_EQ"]))



from .base_panel_learner import BasePanelLearner
from .signal_optimizer import SignalOptimizer
from .beta_estimator import BetaEstimator

__all__ = [
    'BasePanelLearner',
    'SignalOptimizer',
    'BetaEstimator',
]


from .base_regression_system import (
    BaseRegressionSystem,
)

from .regressor_systems import (
    LinearRegressionSystem,
    LADRegressionSystem,
    RidgeRegressionSystem,
    CorrelationVolatilitySystem,
)

__all__ = [
    "BaseRegressionSystem",
    "CorrelationVolatilitySystem",
    "LADRegressionSystem",
    "LinearRegressionSystem",
    "RidgeRegressionSystem",
]


"""
Panel K-Fold cross-validator classes. 
"""

import numpy as np
import pandas as pd

from macrosynergy.learning.splitters.base_splitters import KFoldPanelSplit


class ExpandingKFoldPanelSplit(KFoldPanelSplit):
    """
    Time-respecting K-Fold cross-validator for panel data.

    Parameters
    ----------
    n_splits : int
        Number of folds i.e. (training set, test set) pairs. Default is 5.
        Must be at least 2.

    Notes
    -----
    This splitter can be considered to be a panel data analogue to the `TimeSeriesSplit`
    splitter provided by `scikit-learn`.

    Unique dates in the panel are divided into 'n_splits + 1' sequential and
    non-overlapping intervals, resulting in 'n_splits' pairs of training and test sets.
    The 'i'th training set is the union of the first 'i' intervals, and the 'i'th test set
    is the 'i+1'th interval.
    """

    def _determine_splits(self, unique_dates, n_splits):
        """
        Determine panel time period splits based on the sorted collection of unique dates and the
        number of splits specified by the user.

        Parameters
        ----------
        unique_dates : pd.DatetimeIndex
            Sorted collection of unique dates in the panel.
        n_splits : int
            Number of splits to generate.

        Returns
        -------
        splits : list of np.ndarray
            List of numpy arrays denoting dates in each split.
        """
        return np.array_split(unique_dates, n_splits + 1)

    def _get_split_indicies(self, n_split, splits, Xy, dates, unique_dates):
        """
        Determine the training and test set indices for a given split.

        Parameters
        ----------
        n_split : int
            Index of the current split.
        splits : list of np.ndarray
            List of numpy arrays denoting dates in each split.
        Xy : pd.DataFrame
            Combined dataframe of the features and the target variable.
        dates : pd.DatetimeIndex
            DatetimeIndex of all dates in the panel.
        unique_dates : pd.DatetimeIndex
            Sorted collection of unique dates in the panel.

        Returns
        -------
        train : np.ndarray
            The training set indices for that split.
        test : np.ndarray
            The testing set indices for that split.
        """
        train_split = np.concatenate(splits[: n_split + 1])
        train_indices = np.where(dates.isin(train_split))[0]
        test_indices = np.where(dates.isin(splits[n_split + 1]))[0]

        return train_indices, test_indices


class RollingKFoldPanelSplit(KFoldPanelSplit):
    """
    Unshuffled K-Fold cross-validator for panel data.

    Parameters
    ----------
    n_splits : int
        Number of folds. Default is 5. Must be at least 2.

    Notes
    -----
    This splitter can be considered to be a panel data analogue to the `KFold` splitter
    provided by `scikit-learn`, with `shuffle=False` and with splits determined on the
    time dimension.

    Unique dates in the panel are divided into 'n_splits' sequential and non-overlapping
    intervals of equal size, resulting in 'n_splits' pairs of training and test sets.
    The 'i'th test set is the 'i'th interval, and the 'i'th training set is all other
    intervals.
    """

    def _determine_splits(self, unique_dates, n_splits):
        """
        Determine panel time period splits based on the sorted collection of unique dates and the
        number of splits specified by the user.

        Parameters
        ----------
        unique_dates : pd.DatetimeIndex
            Sorted collection of unique dates in the panel.
        n_splits : int
            Number of splits to generate.

        Returns
        -------
        splits : list of np.ndarray
            List of numpy arrays denoting dates in each split.
        """
        return np.array_split(unique_dates, n_splits)

    def _get_split_indicies(self, n_split, splits, Xy, dates, unique_dates):
        """
        Determine the training and test set indices for a given split.

        Parameters
        ----------
        n_split : int
            Index of the current split.
        splits : list of np.ndarray
            List of numpy arrays denoting dates in each split.
        Xy : pd.DataFrame
            Combined dataframe of the features and the target variable.
        dates : pd.DatetimeIndex
            DatetimeIndex of all dates in the panel.
        unique_dates : pd.DatetimeIndex
            Sorted collection of unique dates in the panel.

        Returns
        -------
        train : np.ndarray
            The training set indices for that split.
        test : np.ndarray
            The testing set indices for that split.
        """
        test_split = splits[n_split]
        train_split = np.concatenate(splits[:n_split] + splits[n_split + 1 :])
        train_indices = np.where(dates.isin(train_split))[0]
        test_indices = np.where(dates.isin(test_split))[0]

        return train_indices, test_indices


class RecencyKFoldPanelSplit(KFoldPanelSplit):
    """
    Time-respecting K-Fold panel cross-validator that creates training and test sets based
    on the most recent samples in the panel.

    Parameters
    ----------
    n_splits : int
        Number of folds i.e. (training set, test set) pairs. Default is 5.
        Must be at least 1.
    n_periods : int
        Number of time periods, in units of native dataset frequency, to comprise each
        test set. Default is 252 (1 year for daily data).

    Notes
    -----
    This splitter is similar to the ExpandingKFoldPanelSplit, except that the sorted
    unique timestamps are not divided into equal intervals. Instead, the last
    `n_periods` * `n_splits` timestamps in the panel are divided into `n_splits`
    non-overlapping intervals, each of which is used as a test set. The corresponding
    training set is comprised of all samples with timestamps earlier than its test set.
    Consequently, this is a K-Fold walk-forward cross-validator, but with test folds
    concentrated on the most recent information.
    """

    def __init__(self, n_splits=5, n_periods=252):
        super().__init__(n_splits=n_splits, min_n_splits=1)

        # Additional checks
        if not isinstance(n_periods, int):
            raise TypeError(f"n_periods must be an integer. Got {type(n_periods)}.")
        if n_periods < 1:
            raise ValueError(
                f"Cannot have number of periods less than 1. Got {n_periods}."
            )

        # Additional attributes
        self.n_periods = n_periods

    def _determine_splits(self, unique_dates, n_splits):
        """
        Determine panel time period splits based on the sorted collection of unique dates and the
        number of splits specified by the user.

        Parameters
        ----------
        unique_dates : pd.DatetimeIndex
            Sorted collection of unique dates in the panel.
        n_splits : int
            Number of splits to generate.

        Returns
        -------
        splits : list of np.ndarray
            List of numpy arrays denoting dates in each split.
        """
        return np.array_split(unique_dates[-n_splits * self.n_periods :], n_splits)

    def _get_split_indicies(self, n_split, splits, Xy, dates, unique_dates):
        """
        Determine the training and test set indices for a given split.

        Parameters
        ----------
        n_split : int
            Index of the current split.
        splits : list of np.ndarray
            List of numpy arrays denoting dates in each split.
        Xy : pd.DataFrame
            Combined dataframe of the features and the target variable.
        dates : pd.DatetimeIndex
            DatetimeIndex of all dates in the panel.
        unique_dates : pd.DatetimeIndex
            Sorted collection of unique dates in the panel.

        Returns
        -------
        train : np.ndarray
            The training set indices for that split.
        test : np.ndarray
            The testing set indices for that split.
        """
        test_split = np.array(splits[n_split], dtype=np.datetime64)
        train_split = unique_dates[unique_dates < test_split[0]]

        train_indices = np.where(dates.isin(train_split))[0]
        test_indices = np.where(dates.isin(test_split))[0]

        return train_indices, test_indices


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf
    import macrosynergy.management as msm

    # Example dataset
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2001-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X = dfd.drop(columns=["XR"])
    y = dfd["XR"]

    """ Single validation set example """
    splitter = RecencyKFoldPanelSplit(n_splits=1, n_periods=21*12)
    splitter.visualise_splits(X, y)

    """ Cross-validation examples """
    # ExpandingKFoldPanelSplit
    splitter = ExpandingKFoldPanelSplit(n_splits=5)
    splitter.visualise_splits(X, y)

    # RollingKFoldPanelSplit
    splitter = RollingKFoldPanelSplit(n_splits=5)
    splitter.visualise_splits(X, y)

    # RecencyKFoldPanelSplit
    splitter = RecencyKFoldPanelSplit(n_splits=4, n_periods=21 * 3)
    splitter.visualise_splits(X, y)



import numpy as np
import pandas as pd

import datetime
from abc import ABC, abstractmethod
from sklearn.base import BaseEstimator, RegressorMixin


class BaseRegressionSystem(BaseEstimator, RegressorMixin, ABC):
    def __init__(
        self,
        roll="full",
        min_xs_samples=2,
        data_freq=None,
    ):
        """
        Base class for systems of regressors.

        Parameters
        ----------
        roll : int or str, default = "full"
            The lookback of the rolling window for the regression. If "full",
            the entire cross-sectional history is used for each regression. Otherwise,
            this should be specified in units of the data frequency. If `data_freq` is not
            None or "unadjusted", then an integer value for `roll` should be expressed in
            units of the data frequency provided in the `data_freq` argument.
        min_xs_samples : int, default=2
            The minimum number of samples required in each cross-section training set for
            a regression model to be fitted for that cross-section. If `data_freq` is None
            or "unadjusted", this parameter is specified in terms of the native dataset
            frequency. Otherwise, this parameter should be expressed in units of the
            frequency specified in the `data_freq` argument.
        data_freq : str, optional
            Training set data frequency. This is primarily to be used within the context
            of market beta estimation in the `BetaEstimator` class in
            `macrosynergy.learning`, allowing one to cross-validate the underlying data
            frequency for good beta estimation. Accepted strings are 'unadjusted', 'W' for
            weekly, 'M' for monthly and 'Q' for quarterly. It is recommended to set this
            parameter to "W", "M" or "Q" only when the native dataset frequency is greater.

        Notes
        -----
        Systems of regressors are used to fit a different regression model on each
        cross-section of a panel. This is useful when one believes the within-group
        relationships are sufficiently different to warrant separate models, or when
        Simpson's paradox is a concern.

        A concern with this approach, however, is that the number of samples in each
        cross-section may be too small to fit a model. This is particularly true when
        dealing with low-frequency macro quantamental data.
        """
        # Checks
        if not isinstance(roll, (str, int)):
            raise TypeError("roll must be either a string or integer.")
        if isinstance(roll, str) and roll != "full":
            raise ValueError("roll must equal `full` when a string is specified.")
        if isinstance(roll, int) and roll <= 1:
            raise ValueError(
                "roll must be greater than 1 when an integer is specified."
            )

        if not isinstance(min_xs_samples, int):
            raise TypeError("The min_xs_samples argument must be an integer.")
        if min_xs_samples < 2:
            raise ValueError("The min_xs_samples argument must be at least 2.")

        if data_freq is not None:
            if not isinstance(data_freq, str):
                raise TypeError("The data_freq argument must be a string.")
            if data_freq not in ["unadjusted", "W", "M", "Q"]:
                raise ValueError(
                    "data_freq must be one of 'unadjusted', 'W', 'M' or 'Q'."
                )

        # Set attributes
        self.roll = roll
        self.data_freq = data_freq
        self.min_xs_samples = min_xs_samples

        self.models_ = None

    def fit(
        self,
        X,
        y,
    ):
        """
        Fit a regression on each cross-section of a panel, subject to availability.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.
        y : pd.Series, pd.DataFrame or np.ndarray
            Target variable.

        Returns
        -------
        self : BaseRegressionSystem
            Fitted regression system object.
        """
        # Checks
        y = self._check_fit_params(X, y)
        self.n_features_in_ = X.shape[1]
        self.feature_names_in_ = X.columns
        self.models_ = {}

        # Downsample data frequency if necessary
        if (self.data_freq is not None) and (self.data_freq != "unadjusted"):
            X = self._downsample_by_data_freq(X)
            y = self._downsample_by_data_freq(y)

        # Iterate over cross-sections and fit a regression model on each
        cross_sections = X.index.unique(level=0)
        for section in cross_sections:
            X_section = X.xs(section, level=0, drop_level=False)
            y_section = y.xs(section, level=0, drop_level=False)

            unique_dates = sorted(X_section.index.unique())
            num_dates = len(unique_dates)
            # Skip cross-sections with insufficient samples
            if not self._check_xs_dates(self.min_xs_samples, num_dates):
                continue
            # Roll the data if necessary
            if self.roll and self.roll != "full":
                if num_dates <= self.roll:
                    continue
                else:
                    X_section, y_section = self.roll_dates(
                        self.roll, X_section, y_section, unique_dates
                    )
            # Fit the model
            self._fit_cross_section(section, X_section, y_section)

        return self

    def _fit_cross_section(self, section, X_section, y_section):
        """
        Fit a regression model on a single cross-section.

        Parameters
        ----------
        section : str
            The identifier of the cross-section.
        X_section : pd.DataFrame
            Input feature matrix for the cross-section.
        y_section : pd.Series
            Target variable for the cross-section.
        """
        model = self.create_model()
        model.fit(pd.DataFrame(X_section), y_section)
        # Store model and coefficients
        self.models_[section] = model
        self.store_model_info(section, model)

    def predict(
        self,
        X,
    ):
        """
        Make predictions over a panel dataset based on trained observation-specific
        models.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.

        Returns
        -------
        predictions : pd.Series
            Pandas series of predictions, multi-indexed by cross-section and date.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError("The X argument must be a pandas DataFrame.")
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not np.all(X.columns == self.feature_names_in_):
            raise ValueError(
                "The input feature matrix must have the same columns as the",
                "training feature matrix.",
            )
        if len(X.columns) != self.n_features_in_:
            raise ValueError(
                "The input feature matrix must have the same number of",
                "columns as the training feature matrix.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix must not contain any missing values."
            )
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for regression systems",
                " must be numeric.",
            )

        predictions = pd.Series(index=X.index, data=np.nan)

        # Store predictions for each test cross-section, if an existing model is available
        cross_sections = predictions.index.get_level_values(0).unique()
        for idx, section in enumerate(cross_sections):
            if section in self.models_.keys():
                # If a model exists, return the estimated OOS contract return.
                predictions[predictions.index.get_level_values(0) == section] = (
                    self.models_[section].predict(X.xs(section, level=0)).flatten()
                )

        return predictions

    def roll_dates(self, roll, X_section, y_section, unique_dates):
        """
        Adjust dataset to be contained within a rolling window.

        Parameters
        ----------
        roll : int
            The lookback of the rolling window.
        X_section : pd.DataFrame
            Input feature matrix for the cross-section.
        y_section : pd.Series
            Target variable for the cross-section.
        unique_dates : list
            List of unique dates in the cross-section.

        Returns
        -------
        X_section : pd.DataFrame
            Input feature matrix for the cross-section, adjusted for the rolling window.
        y_section : pd.Series
            Target variable for the cross-section, adjusted for the rolling window.
        """
        right_dates = unique_dates[-roll:]
        
        common_index = X_section.index.intersection(right_dates)

        X_section = X_section.reindex(common_index)
        y_section = y_section.reindex(common_index)

        return X_section, y_section

    @abstractmethod
    def store_model_info(self, section, model):
        """
        Store necessary model information for explainability.

        Parameters
        ----------
        section : str
            The identifier of the cross-section.
        model : RegressorMixin
            The fitted regression model.

        Notes
        ------
        Must be overridden.
        """
        pass

    @abstractmethod
    def create_model(self):
        """
        Instantiate a regression model for a given cross-section.

        Notes
        -----
        Must be overridden.
        """
        pass

    def _check_xs_dates(self, min_xs_samples, num_dates):
        """
        Cross-sectional availability check.

        Parameters
        ----------
        min_xs_samples : int
            The minimum number of samples required in each cross-section training set for
            a regression model to be fitted.
        num_dates : int
            The number of unique dates in the cross-section.

        Returns
        -------
        bool
            True if the number of samples is sufficient, False otherwise
        """
        if num_dates < min_xs_samples:
            return False

        return True

    def _downsample_by_data_freq(self, df):
        """
        Resample the input dataset to the specified data frequency.

        Parameters
        ----------
        df : pd.DataFrame
            Input feature matrix.

        Returns
        -------
        pd.DataFrame
            Resampled feature matrix.
        """
        return (
            df.groupby(
                [
                    pd.Grouper(level="cid"),
                    pd.Grouper(level="real_date", freq=self.data_freq),
                ]
            )
            .sum()
            .copy()
        )

    def _check_fit_params(self, X, y):
        """
        Input checks for the fit method parameters.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.
        y : pd.Series, pd.DataFrame or np.ndarray
            Target variable.
        """
        # X
        if not isinstance(X, pd.DataFrame):
            raise TypeError("The X argument must be a pandas DataFrame.")
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for regression systems",
                " must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for regression systems must not contain any "
                "missing values."
            )

        if not isinstance(y, (pd.DataFrame, pd.Series, np.ndarray)):
            raise TypeError(
                "The y argument must be a pandas DataFrame, Series or numpy array."
            )
        if len(X) != len(y):
            raise ValueError("The number of samples in X and y must match.")
        if isinstance(y, np.ndarray):
            # This can happen during sklearn's GridSearch when a voting regressor is used
            if y.ndim != 1 and y.ndim != 2:
                raise ValueError("y must be a 1D or 2D array.")
            if y.ndim == 2 and y.shape[1] != 1:
                raise ValueError("y must have only one column.")
            y = pd.Series(y, index=X.index)
        if not isinstance(y, np.ndarray):
            if not np.issubdtype(y.values.dtype, np.number):
                raise ValueError("The target vector must be numeric.")
            if y.isnull().values.any():
                raise ValueError(
                    "The target vector must not contain any missing values."
                )
        else:
            if not np.issubdtype(y.dtype, np.number):
                raise ValueError("The target vector must be numeric.")
            if np.isnan(y).any():
                raise ValueError(
                    "The target vector must not contain any missing values."
                )

        return y



"""
Class to determine and store sequentially-optimized panel forecasts based on statistical
machine learning. 
"""

import numbers

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.feature_selection import SelectorMixin
from sklearn.pipeline import Pipeline

from macrosynergy.learning import ExpandingIncrementPanelSplit
from macrosynergy.learning.sequential import BasePanelLearner
from macrosynergy.management.utils import concat_categorical, _insert_as_categorical
from macrosynergy.management.types import QuantamentalDataFrame


class SignalOptimizer(BasePanelLearner):
    """
    Class for sequential optimization of return forecasts based on panels of quantamental
    features.

    Parameters
    ----------
    df : pd.DataFrame
        Daily quantamental dataframe in JPMaQS format containing a panel of features, as
        well as a panel of returns.
    xcats : list
        List comprising feature names, with the last element being the response variable
        name. The features and the response variable must be categories in the dataframe.
    cids : list, optional
        List of cross-section identifiers for consideration in the panel. Default is None,
        in which case all cross-sections in `df` are considered.
    start : str, optional
        Start date for considered data in subsequent analysis in ISO 8601 format.
        Default is None i.e. the earliest date in the dataframe.
    end : str, optional
        End date for considered data in subsequent analysis in ISO 8601 format.
        Default is None i.e. the latest date in the dataframe.
    blacklist : list, optional
        Blacklisting dictionary specifying date ranges for which cross-sectional
        information should be excluded. The keys are cross-sections and the values
        are tuples of start and end dates in ISO 8601 format. Default is None.
    freq : str, optional
        Frequency of the analysis. Default is "M" for monthly.
    lag : int, optional
        Number of periods to lag the response variable. Default is 1.
    xcat_aggs : list, optional
        List of aggregation functions to apply to the features, used when `freq` is not
        `D`. Default is ["last", "sum"].
    generate_labels : callable, optional
        Function to transform the response variable into either alternative regression
        targets or classification labels. Default is None.

    Notes
    -----
    The `SignalOptimizer` class is used to predict the response variable, usually a panel
    of asset class returns, based on a panel of features that are lagged by a specified
    number of periods. This is done in a sequential manner, by specifying the size of an
    initial training set, choosing an optimal model out of a provided collection
    (with associated hyperparameters), forecasting the return panel, and then expanding
    the training set to include the now-realized returns. The process continues until the
    end of the dataset is reached.

    In addition to storing forecasts, this class also stores useful information for
    analysis such as the models selected at each point in time, the feature coefficients
    and intercepts (where relevant) of selected models, as well as the features
    selected by any feature selection modules.

    Model and hyperparameter selection is performed by cross-validation. Given a
    collection of models and associated hyperparameters to choose from, an HPO is run
    - currently only grid search and random search are supported - to determine the
    optimal choice. This is done by providing a collection of `scikit-learn` compatible
    scoring functions, as well as a collection of `scikit-learn` compatible
    cross-validation splitters and scorers. At each point in time, the cross-validation
    folds are the union of the folds produced by each splitter provided. Each scorer is
    evaluated on each test fold and summarised across test folds by either a custom
    function provided by the user or a common string i.e. 'mean'.

    Consequently, each model and hyperparameter combination has an associated collection
    of scores induced by different metrics, in units of those scorers. In order to form a
    composite score for each hyperparameter, the scores must be normalized across
    model/hyperparameter combinations. This makes scores across scorers comparable, so
    that the average score across adjusted scores can be used as a meaningful estimate
    of each model's generalization ability. Finally, a composite score for each model and
    hyperparameter combination is calculated by averaging the adjusted scores across all
    scorers.

    The optimal model is the one with the largest composite score.
    """

    def __init__(
        self,
        df,
        xcats,
        cids=None,
        start=None,
        end=None,
        blacklist=None,
        freq="M",
        lag=1,
        xcat_aggs=["last", "sum"],
        generate_labels=None,
    ):
        # Run checks and necessary dataframe massaging
        super().__init__(
            df=df,
            xcats=xcats,
            cids=cids,
            start=start,
            end=end,
            blacklist=blacklist,
            freq=freq,
            lag=lag,
            xcat_aggs=xcat_aggs,
            generate_labels=generate_labels,
        )

        # Create forecast dataframe index
        min_date = min(self.unique_date_levels)
        max_date = max(self.unique_date_levels)
        forecast_date_levels = pd.date_range(start=min_date, end=max_date, freq="B")
        self.forecast_idxs = pd.MultiIndex.from_product(
            [self.unique_xs_levels, forecast_date_levels], names=["cid", "real_date"]
        )

        # Create initial dataframes to store relevant quantities from the learning process
        self.preds = pd.DataFrame(columns=["real_date", "cid", "xcat", "value"]).astype(
            {
                "real_date": "datetime64[ns]",
                "cid": "category",
                "xcat": "category",
                "value": "float32",
            }
        )
        self.feature_importances = pd.DataFrame(
            columns=["real_date", "name"] + list(self.X.columns)
        ).astype(
            {
                **{col: "float32" for col in self.X.columns},
                "real_date": "datetime64[ns]",
                "name": "category",
            }
        )
        self.intercepts = pd.DataFrame(
            columns=["real_date", "name", "intercepts"]
        ).astype(
            {
                "real_date": "datetime64[ns]",
                "name": "category",
                "intercepts": "float32",
            }
        )

        self.selected_ftrs = pd.DataFrame(
            columns=["real_date", "name"] + list(self.X.columns)
        ).astype(
            {
                **{col: "int" for col in self.X.columns},
                "real_date": "datetime64[ns]",
                "name": "category",
            }
        )

        self.store_correlations = False
        # Create data structure to store correlation matrix of features feeding into the
        # final model and the input features themselves
        self.ftr_corr = pd.DataFrame(
            columns=[
                "real_date",
                "name",
                "predictor_input",
                "pipeline_input",
                "pearson",
            ]
        ).astype(
            {
                "real_date": "datetime64[ns]",
                "name": "category",
                "predictor_input": "category",
                "pipeline_input": "category",
                "pearson": "float",
            }
        )

    def calculate_predictions(
        self,
        name,
        models,
        hyperparameters,
        scorers,
        inner_splitters,
        search_type="grid",
        normalize_fold_results=False,
        cv_summary="mean",
        min_cids=4,
        min_periods=12 * 3,
        test_size=1,
        max_periods=None,
        split_functions=None,
        n_iter=None,
        n_jobs_outer=-1,
        n_jobs_inner=1,
        store_correlations=False,
    ):
        """
        Determine forecasts and store relevant quantities over time.

        Parameters
        ----------
        name : str
            Name of the signal optimization process.
        models : dict
            Dictionary of models to choose from. The keys are model names and the values
            are scikit-learn compatible models.
        hyperparameters : dict
            Dictionary of hyperparameters to choose from. The keys are model names and
            the values are hyperparameter dictionaries for the corresponding model. The
            keys must match with those provided in `models`.
        scorers : dict
            Dictionary of scoring functions to use in the hyperparameter optimization
            process. The keys are scorer names and the values are scikit-learn compatible
            scoring functions.
        inner_splitters : dict
            Dictionary of inner splitters to use in the hyperparameter optimization
            process. The keys are splitter names and the values are scikit-learn compatible
            cross-validator objects.
        search_type : str, optional
            Type of hyperparameter optimization to perform. Default is "grid". Options are
            "grid" and "prior".
        normalize_fold_results : bool, optional
            Whether to normalize the scores across folds before combining them. Default is
            False.
        cv_summary : str or callable, optional
            Summary function to use to combine scores across cross-validation folds.
            Default is "mean". Options are "mean", "median" or a callable function.
        min_cids : int, optional
            Minimum number of cross-sections required for the initial
            training set. Default is 4.
        min_periods : int, optional
            Minimum number of periods required for the initial training set, in units of
            the frequency `freq` specified in the constructor. Default is 36.
        test_size : int, optional
            Number of periods to pass before retraining a selected model. Default is 1.
        max_periods : int, optional
            Maximum length of each training set in units of the frequency `freq` specified
            in the constructor. Default is None, in which case the sequential optimization
            uses expanding training sets, as opposed to rolling windows.
        split_functions : dict, optional
            Dict of callables for determining the number of cross-validation
            splits to add to the initial number as a function of the number of iterations
            passed in the sequential learning process. Default is None. The keys must
            correspond to the keys in `inner_splitters` and should be set to None for any
            splitters that do not require splitter adjustment.
        n_iter : int, optional
            Number of iterations to run in random hyperparameter search. Default is None.
        n_jobs_outer : int, optional
            Number of jobs to run in parallel for the outer sequential loop. Default is -1.
            It is advised for n_jobs_inner * n_jobs_outer (replacing -1 with the number of
            available cores) to be less than or equal to the number of available cores on
            the machine.
        n_jobs_inner : int, optional
            Number of jobs to run in parallel for the inner loop. Default is 1.
            It is advised for n_jobs_inner * n_jobs_outer (replacing -1 with the number of
            available cores) to be less than or equal to the number of available cores on
            the machine.
        store_correlations : bool
            Whether to store the correlations between input pipeline features and input
            predictor features. Default is False.
        """
        if not isinstance(store_correlations, bool):
            raise TypeError("The store_correlations argument must be a boolean.")

        if store_correlations and not all(
            [isinstance(model, Pipeline) for model in models.values()]
        ):
            raise ValueError(
                "The store_correlations argument is only valid when all models are Scikit-learn Pipelines."
            )
        self.store_correlations = store_correlations

        # Set up outer splitter
        outer_splitter = ExpandingIncrementPanelSplit(
            train_intervals=test_size,
            test_size=test_size,
            min_cids=min_cids,
            min_periods=min_periods,
            max_periods=max_periods,
        )

        results = self.run(
            name=name,
            outer_splitter=outer_splitter,
            inner_splitters=inner_splitters,
            models=models,
            hyperparameters=hyperparameters,
            scorers=scorers,
            search_type=search_type,
            normalize_fold_results=normalize_fold_results,
            cv_summary=cv_summary,
            split_functions=split_functions,
            n_iter=n_iter,
            n_jobs_outer=n_jobs_outer,
            n_jobs_inner=n_jobs_inner,
        )

        self._check_duplicate_results(name)

        # Collect results from the worker
        # quantamental_data, model_data, other_data
        prediction_data = []
        model_choice_data = []
        ftr_coef_data = []
        intercept_data = []
        ftr_selection_data = []
        ftr_corr_data = []

        for split_result in results:
            prediction_data.append(split_result["predictions"])
            model_choice_data.append(split_result["model_choice"])
            ftr_coef_data.append(split_result["feature_importances"])
            intercept_data.append(split_result["intercepts"])
            ftr_selection_data.append(split_result["selected_ftrs"])
            ftr_corr_data.extend(split_result["ftr_corr"])

        # First create pandas dataframes to store the forecasts
        forecasts_df = pd.DataFrame(
            index=self.forecast_idxs, columns=[name], data=np.nan, dtype="float32"
        )
        # Create quantamental dataframe of forecasts
        for idx, forecasts in prediction_data:
            forecasts_df.loc[idx, name] = forecasts

        forecasts_df = forecasts_df.groupby(level=0).ffill().dropna()

        if self.blacklist is not None:
            for cross_section, periods in self.blacklist.items():
                cross_section_key = cross_section.split("_")[0]
                if cross_section_key in self.unique_xs_levels:
                    forecasts_df.loc[
                        (cross_section_key, slice(periods[0], periods[1])), :
                    ] = np.nan

        forecasts_df.columns = forecasts_df.columns.astype("category")
        forecasts_df_long = pd.melt(
            frame=forecasts_df.reset_index(),
            id_vars=["real_date", "cid"],
            var_name="xcat",
        )
        self.preds = concat_categorical(
            df1=self.preds,
            df2=forecasts_df_long,
        )

        # Store model selection data
        model_df_long = pd.DataFrame(
            columns=[col for col in self.chosen_models.columns if col != "name"],
            data=model_choice_data,
        ).astype({"model_type": "category"})
        model_df_long = _insert_as_categorical(model_df_long, "name", name, 1)

        self.chosen_models = concat_categorical(
            df1=self.chosen_models,
            df2=model_df_long,
        )

        # Store feature coefficients
        coef_df_long = pd.DataFrame(
            columns=[col for col in self.feature_importances.columns if col != "name"],
            data=ftr_coef_data,
        )
        coef_df_long = _insert_as_categorical(coef_df_long, "name", name, 1)
        self.feature_importances = concat_categorical(
            self.feature_importances,
            coef_df_long,
        )

        # Store intercept
        intercept_df_long = pd.DataFrame(
            columns=[col for col in self.intercepts.columns if col != "name"],
            data=intercept_data,
        )
        intercept_df_long = _insert_as_categorical(intercept_df_long, "name", name, 1)
        self.intercepts = concat_categorical(
            self.intercepts,
            intercept_df_long,
        )

        # Store selected features
        ftr_select_df_long = pd.DataFrame(
            columns=[col for col in self.selected_ftrs.columns if col != "name"],
            data=ftr_selection_data,
        )
        ftr_select_df_long = _insert_as_categorical(ftr_select_df_long, "name", name, 1)
        self.selected_ftrs = concat_categorical(
            self.selected_ftrs,
            ftr_select_df_long,
        )

        ftr_corr_df_long = pd.DataFrame(
            columns=self.ftr_corr.columns, data=ftr_corr_data
        )

        self.ftr_corr = concat_categorical(
            self.ftr_corr,
            ftr_corr_df_long,
        )

    def _check_duplicate_results(self, name):
        conditions = [
            ("preds", "xcat", name),
            ("feature_importances", "name", name),
            ("intercepts", "name", name),
            ("selected_ftrs", "name", name),
            ("ftr_corr", "name", name),
            ("chosen_models", "name", name),
        ]
        self._remove_results(conditions)

    def store_split_data(
        self,
        pipeline_name,
        optimal_model,
        optimal_model_name,
        optimal_model_score,
        optimal_model_params,
        inner_splitters_adj,
        X_train,
        y_train,
        X_test,
        y_test,
        timestamp,
        adjusted_test_index,
    ):
        """
        Stores characteristics of the optimal model at each retraining date.

        Parameters
        ----------
        pipeline_name : str
            Name of the signal optimization process.
        optimal_model : RegressorMixin, ClassifierMixin or Pipeline
            Optimal model selected at each retraining date.
        optimal_model_name : str
            Name of the optimal model.
        optimal_model_score : float
            Cross-validation score for the optimal model.
        optimal_model_params : dict
            Chosen hyperparameters for the optimal model.
        inner_splitters_adj : dict
            Dictionary of adjusted inner splitters.
        X_train : pd.DataFrame
            Training feature matrix.
        y_train : pd.Series
            Training response variable.
        X_test : pd.DataFrame
            Test feature matrix.
        y_test : pd.Series
            Test response variable.
        timestamp : pd.Timestamp
            Timestamp of the retraining date.
        adjusted_test_index : pd.MultiIndex
            Adjusted test index to account for lagged features.

        Returns
        -------
        dict
            Dictionary containing feature importance scores, intercepts, selected features
            and correlations between inputs to pipelines and those entered into a final
            model.
        """
        if optimal_model is not None:
            if hasattr(optimal_model, "create_signal"):
                if callable(getattr(optimal_model, "create_signal")):
                    preds = optimal_model.create_signal(X_test)
            else:
                preds = optimal_model.predict(X_test)
        else:
            preds = np.zeros(X_test.shape[0])

        prediction_data = [adjusted_test_index, preds]

        feature_names = np.array(X_train.columns)
        if isinstance(optimal_model, Pipeline):
            final_estimator = optimal_model[-1]
            for _, transformer in reversed(optimal_model.steps):
                if isinstance(transformer, SelectorMixin):
                    feature_names = transformer.get_feature_names_out()
                    break
        else:
            final_estimator = optimal_model

        coefs = np.full(X_train.shape[1], np.nan)

        if hasattr(final_estimator, "feature_importances_"):
            coef = final_estimator.feature_importances_
        elif hasattr(final_estimator, "coef_"):
            coef = final_estimator.coef_

        if coef.ndim == 1:
            coefs = coef
        elif coef.ndim == 2:
            if coef.shape[0] == 1:
                coefs = coef.flatten()

        coef_ftr_map = {ftr: coef for ftr, coef in zip(feature_names, coefs)}
        coefs = [
            coef_ftr_map[ftr] if ftr in coef_ftr_map else np.nan
            for ftr in X_train.columns
        ]
        if hasattr(final_estimator, "intercept_"):
            if isinstance(final_estimator.intercept_, np.ndarray):
                # Store the intercept if it has length one
                if len(final_estimator.intercept_) == 1:
                    intercepts = final_estimator.intercept_[0]
                else:
                    intercepts = np.nan
            else:
                # The intercept will be a float/integer
                intercepts = final_estimator.intercept_
        else:
            intercepts = np.nan

        # Get feature selection information
        if len(feature_names) == X_train.shape[1]:
            # Then all features were selected
            ftr_selection_data = [timestamp] + [1 for _ in feature_names]
        else:
            # Then some features were excluded
            ftr_selection_data = [timestamp] + [
                1 if name in feature_names else 0 for name in np.array(X_train.columns)
            ]

        ftr_corr_data = self._get_ftr_corr_data(
            pipeline_name, optimal_model, X_train, timestamp
        )

        # Store data
        split_result = {
            "feature_importances": [timestamp] + coefs,
            "intercepts": [timestamp, intercepts],
            "selected_ftrs": ftr_selection_data,
            "predictions": prediction_data,
            "ftr_corr": ftr_corr_data,
        }

        return split_result

    def _get_ftr_corr_data(self, pipeline_name, optimal_model, X_train, timestamp):
        """
        Returns a list of correlations between the input features to a pipeline and the
        features inputted into the final model, at each retraining date.

        Parameters
        ----------
        pipeline_name : str
            Name of the signal optimization process.
        optimal_model : RegressorMixin, ClassifierMixin or Pipeline
            Optimal model selected at each retraining date.
        X_train : pd.DataFrame
            Input feature matrix.
        timestamp : pd.Timestamp
            Timestamp of the retraining date.

        Returns
        -------
        list
            List of correlations between the input features to a pipeline and the
            features inputted into the final model, at each retraining date.
        """
        if self.store_correlations and optimal_model is not None:
            # Transform the training data to the final feature space
            transformers = Pipeline(steps=optimal_model.steps[:-1])
            X_train_transformed = transformers.transform(X_train)
            n_features = X_train_transformed.shape[1]
            feature_names = (
                X_train_transformed.columns
                if isinstance(X_train_transformed, pd.DataFrame)
                else [f"Feature {i+1}" for i in range(n_features)]
            )
            # Calculate correlation between each original feature in X_train and
            # the transformed features in X_train_transformed
            if isinstance(X_train_transformed, pd.DataFrame):
                X_train_transformed = X_train_transformed.values

            ftr_corr_data = [
                [
                    timestamp,
                    pipeline_name,
                    final_feature_name,
                    input_feature_name,
                    np.corrcoef(
                        X_train_transformed[:, idx],
                        X_train[input_feature_name],
                    )[0, 1],
                ]
                for idx, final_feature_name in enumerate(feature_names)
                for input_feature_name in X_train.columns
            ]

        elif self.store_correlations and optimal_model is None:
            ftr_corr_data = [
                [
                    timestamp,
                    pipeline_name,
                    feature_name,
                    feature_name,
                    1,
                ]
                for feature_name in X_train.columns
            ]
        else:
            ftr_corr_data = []

        return ftr_corr_data

    def get_optimized_signals(self, name=None):
        """
        Returns optimized signals for one or more processes

        Parameters
        ----------
        name : str or list, optional
            Label(s) of signal optimization process(es). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe in JPMaQS format of working daily predictions.
        """
        if name is None:
            preds = self.preds
        else:
            if isinstance(name, str):
                name = [name]
            elif not isinstance(name, list):
                raise TypeError(
                    "The process name must be a string or a list of strings."
                )

            for n in name:
                if n not in self.preds.xcat.unique():
                    raise ValueError(
                        f"""The process name '{n}' is not in the list of already-run
                        pipelines. Please check the name carefully. If correct, please run 
                        calculate_predictions() first.
                        """
                    )
            preds = self.preds[self.preds.xcat.isin(name)]
        # return self.preds[self.preds.xcat.isin(name)]
        signals_df = QuantamentalDataFrame(
            df=preds,
            categorical=self.df.InitializedAsCategorical,
        ).to_original_dtypes()
        return signals_df

    def get_selected_features(self, name=None):
        """
        Returns the selected features over time for one or more processes.

        Parameters
        ----------
        name: str or list, optional
            Label(s) of signal optimization process(es). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the selected features at each retraining date.
        """
        if name is None:
            return self.selected_ftrs
        else:
            if isinstance(name, str):
                name = [name]
            elif not isinstance(name, list):
                raise TypeError(
                    "The process name must be a string or a list of strings."
                )

            for n in name:
                if n not in self.selected_ftrs.name.unique():
                    raise ValueError(
                        f"""The process name '{n}' is not in the list of already-run
                        pipelines. Please check the name carefully. If correct, please run 
                        calculate_predictions() first.
                        """
                    )
            return self.selected_ftrs[self.selected_ftrs.name.isin(name)]

    def get_feature_importances(self, name=None):
        """
        Returns feature importances for a given pipeline.

        Parameters
        ----------
        name: str or list, optional
            Label(s) of signal optimization process(es). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the feature importances, if available, learnt at each
            retraining date for a given pipeline.

        Notes
        -----
        Availability of feature importances is subject to the selected model having a 
        `feature_importances_` or `coef_` attribute.
        """
        if name is None:
            return self.feature_importances
        else:
            if isinstance(name, str):
                name = [name]
            elif not isinstance(name, list):
                raise TypeError(
                    "The process name must be a string or a list of strings."
                )

            for n in name:
                if n not in self.feature_importances.name.unique():
                    raise ValueError(
                        f"""The process name '{n}' is not in the list of already-run
                        pipelines. Please check the name carefully. If correct, please run 
                        calculate_predictions() first.
                        """
                    )
            return self.feature_importances[
                self.feature_importances.name.isin(name)
            ].sort_values(by="real_date")

    def get_intercepts(self, name=None):
        """
        Returns intercepts for a given pipeline.

        Parameters
        ----------
        name: str or list, optional
            Label(s) of signal optimization process(es). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the intercepts, if available, learnt at each retraining
            date for a given pipeline.
        """
        if name is None:
            return self.intercepts
        else:
            if isinstance(name, str):
                name = [name]
            elif not isinstance(name, list):
                raise TypeError(
                    "The process name must be a string or a list of strings."
                )

            for n in name:
                if n not in self.intercepts.name.unique():
                    raise ValueError(
                        f"""The process name '{n}' is not in the list of already-run
                        pipelines. Please check the name carefully. If correct, please run 
                        calculate_predictions() first.
                        """
                    )
            return self.intercepts[self.intercepts.name.isin(name)].sort_values(
                by="real_date"
            )

    def get_feature_correlations(
        self,
        name=None,
    ):
        """
        Returns dataframe of feature correlations for one or more processes

        Parameters
        ----------
        name: str or list, optional
            Label(s) of signal optimization process(es). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the correlations between the features passed into a model
            pipeline and the post-processed features inputted into the final model.
        """
        if name is None:
            return self.ftr_corr
        else:
            if isinstance(name, str):
                name = [name]
            elif not isinstance(name, list):
                raise TypeError(
                    "The process name must be a string or a list of strings."
                )

            for n in name:
                if n not in self.ftr_corr.name.unique():
                    raise ValueError(
                        f"""Either the process name '{n}' is not in the list of already-run
                        pipelines, or no correlations were stored for this pipeline.
                        Please check the name carefully. If correct, please run 
                        calculate_predictions() first.
                        """
                    )
            return self.ftr_corr[self.ftr_corr.name.isin(name)]

    def feature_selection_heatmap(
        self,
        name,
        remove_blanks=True,
        title=None,
        cap=None,
        ftrs_renamed=None,
        figsize=(12, 8),
    ):
        """
        Visualise the features chosen by the final selector in a scikit-learn pipeline
        over time, for a given signal optimization process that has been run.

        Parameters
        ----------
        name : str
            Name of the previously run signal optimization process.
        remove_blanks : bool, optional
            Whether to remove features from the heatmap that were never selected. Default
            is True.
        title : str, optional
            Title of the heatmap. Default is None. This creates a figure title of the form
            "Model Selection Heatmap for {name}".
        cap : int, optional
            Maximum number of features to display. Default is None. The chosen features
            are the 'cap' most frequently occurring in the pipeline.
        ftrs_renamed : dict, optional
            Dictionary to rename the feature names for visualisation in the plot axis.
            Default is None, which uses the original feature names.
        figsize : tuple of floats or ints, optional
            Tuple of floats or ints denoting the figure size. Default is (12, 8).

        Notes
        -----
        This method displays the features selected by the final selector in a scikit-learn
        pipeline over time, for a given signal optimization process that has been run.
        This information is contained within a binary heatmap.
        This does not take into account inherent feature selection within the predictor.
        """
        # Checks
        self._checks_feature_selection_heatmap(
            name=name, title=title, ftrs_renamed=ftrs_renamed, figsize=figsize
        )

        # Get the selected features for the specified pipeline to visualise selection.
        selected_ftrs = self.get_selected_features(name=name)
        selected_ftrs["real_date"] = selected_ftrs["real_date"].dt.date
        selected_ftrs = (
            selected_ftrs.sort_values(by="real_date")
            .drop(columns=["name"])
            .set_index("real_date")
        )

        # Sort dataframe columns in descending order of the number of times they were selected
        ftr_count = selected_ftrs.sum().sort_values(ascending=False)
        if remove_blanks:
            ftr_count = ftr_count[ftr_count > 0]
        if cap is not None:
            ftr_count = ftr_count.head(cap)

        reindexed_columns = ftr_count.index
        selected_ftrs = selected_ftrs[reindexed_columns]
        if ftrs_renamed is not None:
            selected_ftrs.rename(columns=ftrs_renamed, inplace=True)

        # Create the heatmap
        plt.figure(figsize=figsize)
        if np.all(selected_ftrs == 1):
            sns.heatmap(selected_ftrs.T, cmap="binary_r", cbar=False)
        else:
            sns.heatmap(selected_ftrs.T, cmap="binary", cbar=False)
        plt.title(title)
        plt.show()

    def _checks_feature_selection_heatmap(
        self,
        name: str,
        title=None,
        ftrs_renamed=None,
        figsize=(12, 8),
    ):
        """
        Checks for the feature_selection_heatmap method.

        Parameters
        ----------
        name : str
            Name of the previously run signal optimization process.
        title : str, optional
            Title of the heatmap. Default is None. This creates a figure title of the form
            "Feature Selection Heatmap for {name}".
        ftrs_renamed : dict, optional
            Dictionary to rename the feature names for visualisation in the plot axis.
            Default is None, which uses the original feature names.
        figsize : tuple of floats or ints, optional
            Tuple of floats or ints denoting the figure size. Default is (12, 8).
        """
        if not isinstance(name, str):
            raise TypeError("The pipeline name must be a string.")
        if name not in self.selected_ftrs.name.unique():
            raise ValueError(
                f"""The pipeline name {name} is not in the list of already-calculated 
                pipelines. Please check the pipeline name carefully. If correct, please 
                run calculate_predictions() first.
                """
            )
        if title is None:
            title = f"Feature Selection Heatmap for {name}"
        if not isinstance(title, str):
            raise TypeError("The figure title must be a string.")
        if not isinstance(figsize, tuple):
            raise TypeError("The figsize argument must be a tuple.")
        if len(figsize) != 2:
            raise ValueError("The figsize argument must be a tuple of length 2.")
        for element in figsize:
            if not isinstance(element, (int, float)):
                raise TypeError(
                    "The elements of the figsize tuple must be floats or ints."
                )
        if ftrs_renamed is not None:
            if not isinstance(ftrs_renamed, dict):
                raise TypeError("The ftrs_renamed argument must be a dictionary.")
            for key, value in ftrs_renamed.items():
                if not isinstance(key, str):
                    raise TypeError(
                        "The keys of the ftrs_renamed dictionary must be strings."
                    )
                if not isinstance(value, str):
                    raise TypeError(
                        "The values of the ftrs_renamed dictionary must be strings."
                    )
                if key not in self.X.columns:
                    raise ValueError(
                        f"""The key {key} in the ftrs_renamed dictionary is not a feature 
                        in the pipeline {name}.
                        """
                    )

    def correlations_heatmap(
        self,
        name: str,
        feature_name: str,
        title: str = None,
        cap: int = None,
        ftrs_renamed: dict = None,
        figsize: tuple = (12, 8),
    ):
        """
        Method to visualise correlations between features entering a model, and those that
        entered a preprocessing pipeline.

        Parameters
        ----------
        name : str
            Name of the signal optimization process.
        feature_name : str
            Name of the feature passed into the final predictor.
        title : str, optional
            Title of the heatmap. Default is None. This creates a figure title of the form
            "Correlation Heatmap for feature {feature_name} and pipeline {name}".
        cap : int, optional
            Maximum number of correlations to display. Default is None. The chosen features
            are the 'cap' most highly correlated.
        ftrs_renamed : dict, optional
            Dictionary to rename the feature names for visualisation in the plot axis.
            Default is None, which uses the original feature names.
        figsize : tuple of floats or ints, optional
            Tuple of floats or ints denoting the figure size. Default is (12, 8).

        Notes
        -----
        This method displays the correlation between a feature that is about to be entered
        into a final predictor and the `cap` most correlated features entered into the
        original pipeline. This information is contained within a heatmap.

        In a given pipeline, the features that enter it can be transformed in any way.
        Sometimes the transformation is non-trivial, resulting in a feature space that is
        not easily interpretable. This method allows the user to see how the original
        features are correlated with the features that enter the final model, providing
        insight into the transformation process.

        As an example, dimensionality reduction techniques such as PCA and LDA rotate the
        feature space, resulting in factors that can be hard to interpret. A neural network
        aims to learn a non-linear transformation of the feature space, which can also be
        hard to interpret. This method allows the user to see how the original features are
        correlated with the transformed features, providing insight into the transformation
        that took place.
        """
        # Checks
        self._checks_correlations_heatmap(
            name=name,
            feature_name=feature_name,
            title=title,
            cap=cap,
            ftrs_renamed=ftrs_renamed,
            figsize=figsize,
        )

        # Get the correlations
        correlations = self.get_feature_correlations(name=name)
        correlations = correlations[correlations.predictor_input == feature_name]
        correlations = correlations.sort_values(by="real_date").drop(columns=["name"])
        correlations["real_date"] = correlations["real_date"].dt.date

        # Sort this dataframe based on the average correlation with each feature in
        # pipeline_input
        avg_corr = correlations.groupby("pipeline_input", observed=True)[
            "pearson"
        ].mean()
        avg_corr = avg_corr.sort_values(ascending=False)
        if cap is not None:
            avg_corr = avg_corr.head(cap)

        reindexed_columns = avg_corr.index
        correlations = correlations[correlations.pipeline_input.isin(reindexed_columns)]
        if ftrs_renamed is not None:
            # rename items in correlations.pipeline_input based on ftrs_renamed
            # but leave items not in ftrs_renamed as they are
            correlations["pipeline_input"] = correlations["pipeline_input"].map(
                lambda x: ftrs_renamed.get(x, x)
            )

        # Create the heatmap
        plt.figure(figsize=figsize)
        sns.heatmap(
            correlations.pivot(
                index="pipeline_input", columns="real_date", values="pearson"
            ),
            cmap="coolwarm_r",
            cbar=True,
        )
        if title is None:
            title = (
                f"Correlation Heatmap for feature {feature_name} and pipeline {name}"
            )
        plt.title(title)
        plt.show()

    def _checks_correlations_heatmap(
        self,
        name: str,
        feature_name: str,
        title: str,
        cap: int,
        ftrs_renamed: dict,
        figsize: tuple,
    ):
        """
        Checks for the correlations_heatmap method.

        Parameters
        ----------
        name : str
            Name of the signal optimization process.
        feature_name : str
            Name of the feature passed into the final predictor.
        title : str
            Title of the heatmap. Default is None. This creates a figure title of the form
            "Correlation Heatmap for feature {feature_name} and pipeline {name}".
        cap : int
            Maximum number of correlations to display. Default is None. The chosen features
            are the 'cap' most highly correlated.
        ftrs_renamed : dict
            Dictionary to rename the feature names for visualisation in the plot axis.
            Default is None, which uses the original feature names.
        figsize : tuple of floats or ints
            Tuple of floats or ints denoting the figure size. Default is (12, 8).
        """
        # name
        if not isinstance(name, str):
            raise TypeError("The pipeline name must be a string.")
        if name not in self.ftr_corr.name.unique():
            raise ValueError(
                f"""The pipeline name {name} is not in the list of pipelines with calculated
                correlation matrices. Please check the pipeline name carefully. If correct, please 
                run calculate_predictions() first, or make sure `store_correlations` is
                turned on. 
                """
            )
        # feature name
        if not isinstance(feature_name, str):
            raise TypeError("The feature name must be a string.")
        if feature_name not in self.ftr_corr.predictor_input.unique():
            raise ValueError(
                f"""The feature name {feature_name} is not in the list of features that
                were passed into the final predictor. Please check the feature name carefully.
                """
            )
        # title
        if title is not None:
            if not isinstance(title, str):
                raise TypeError("The title must be a string.")
        # cap
        if cap is not None:
            if not isinstance(cap, int):
                raise TypeError("The cap must be an integer.")
            if cap <= 0:
                raise ValueError("The cap must be greater than zero.")
        # ftrs_renamed
        if ftrs_renamed is not None:
            if not isinstance(ftrs_renamed, dict):
                raise TypeError("The ftrs_renamed argument must be a dictionary.")
            for key, value in ftrs_renamed.items():
                if not isinstance(key, str):
                    raise TypeError(
                        "The keys of the ftrs_renamed dictionary must be strings."
                    )
                if not isinstance(value, str):
                    raise TypeError(
                        "The values of the ftrs_renamed dictionary must be strings."
                    )
                if key not in self.X.columns:
                    raise ValueError(
                        f"""The key {key} in the ftrs_renamed dictionary is not a feature 
                        in the pipeline {name}.
                        """
                    )
        # figsize
        if not isinstance(figsize, tuple):
            raise TypeError("The figsize argument must be a tuple.")
        if len(figsize) != 2:
            raise ValueError("The figsize argument must be a tuple of length 2.")
        for element in figsize:
            if not isinstance(element, numbers.Number) or isinstance(element, bool):
                raise TypeError(
                    "The elements of the figsize tuple must be floats or ints."
                )

    def feature_importance_timeplot(
        self,
        name,
        ftrs=None,
        title=None,
        ftrs_renamed=None,
        figsize=(10, 6),
    ):
        """
        Visualise time series of feature importances for the final predictor in a
        given pipeline, when available. 

        Parameters
        ----------
        name : str
            Name of the previously run signal optimization process.
        ftrs : list, optional
            List of feature names to plot. Default is None.
        title : str, optional
            Title of the plot. Default is None. This creates a figure title of the form
            "Feature importances for pipeline: {name}".
        ftrs_renamed : dict, optional
            Dictionary to rename the feature names for visualisation in the plot legend.
            Default is None, which uses the original feature names.
        figsize : tuple of floats or ints, optional
            Tuple of floats or ints denoting the figure size. Default is (10, 6).

        Notes
        -----
        This method displays the time series of feature importances for a given pipeline,
        when available. Availability depends on whether or not the final predictor in the
        pipeline has either a `coefs_` or `feature_importances_` attribute. This
        information is contained within a line plot. The default behaviour is to sort the
        feature importance columns in ascending order of the number of NAs, accounting for
        a possible feature selection module in the pipeline and plot the feature
        importances for the first 10 features in the sorted order. If more than 10
        features were involved in the learning procedure, the default is to plot the
        feature importances for the first 10 sorted features. By specifying a `ftrs` list
        (which can be no longer than 10 elements in length), this default behaviour can be
        overridden.

        By sorting by NAs, the plot displays the model feature importances for either the
        first 10 features in the dataframe or, when a feature selection module was present,
        the 10 most frequently selected features. 
        """
        # Checks
        if not isinstance(name, str):
            raise TypeError("The pipeline name must be a string.")
        if name not in self.feature_importances.name.unique():
            raise ValueError(
                f"""The pipeline name {name} is not in the list of already-calculated 
                pipelines. Please check the pipeline name carefully. If correct, please 
                run calculate_predictions() first.
                """
            )
        ftrcoef_df = self.get_feature_importances(name)
        if ftrcoef_df.iloc[:, 2:].isna().all().all():
            raise ValueError(
                f"""There are no non-NA feature importances for the pipeline {name}.
                Cannot display a time series plot.
                """
            )
        if ftrs is not None:
            if not isinstance(ftrs, list):
                raise TypeError("The ftrs argument must be a list.")
            if len(ftrs) > 10:
                raise ValueError(
                    "The ftrs list must be no longer than 10 elements in length."
                )
            for ftr in ftrs:
                if not isinstance(ftr, str):
                    raise TypeError("The elements of the ftrs list must be strings.")
                if ftr not in ftrcoef_df.columns:
                    raise ValueError(
                        f"""The feature {ftr} is not in the list of feature coefficients 
                        for the pipeline {name}.
                        """
                    )
        if not isinstance(title, str) and title is not None:
            raise TypeError("The title must be a string.")
        if ftrs_renamed is not None:
            if not isinstance(ftrs_renamed, dict):
                raise TypeError("The ftrs_renamed argument must be a dictionary.")
            for key, value in ftrs_renamed.items():
                if not isinstance(key, str):
                    raise TypeError(
                        "The keys of the ftrs_renamed dictionary must be strings."
                    )
                if not isinstance(value, str):
                    raise TypeError(
                        "The values of the ftrs_renamed dictionary must be strings."
                    )
                if key not in self.X.columns:
                    raise ValueError(
                        f"""The key {key} in the ftrs_renamed dictionary is not a feature 
                        in the pipeline {name}.
                        """
                    )
        if not isinstance(figsize, tuple):
            raise TypeError("The figsize argument must be a tuple.")
        if len(figsize) != 2:
            raise ValueError("The figsize argument must be a tuple of length 2.")
        for element in figsize:
            if not isinstance(element, (int, float, np.int_, np.float_)):
                raise TypeError(
                    "The elements of the figsize tuple must be floats or ints."
                )

        # Set the style
        sns.set_style("darkgrid")

        # Reshape dataframe for plotting
        ftrcoef_df = self.get_feature_importances(name)
        ftrcoef_df = ftrcoef_df.set_index("real_date")
        ftrcoef_df = ftrcoef_df.iloc[:, 1:]

        # Sort dataframe columns in ascending order of the number of Na values in the columns
        na_count = ftrcoef_df.isna().sum().sort_values()
        reindexed_columns = na_count.index
        ftrcoef_df = ftrcoef_df[reindexed_columns]

        if ftrs is not None:
            ftrcoef_df = ftrcoef_df[ftrs]
        else:
            if ftrcoef_df.shape[1] > 11:
                ftrcoef_df = pd.concat(
                    (ftrcoef_df.iloc[:, :10], ftrcoef_df.iloc[:, -1]), axis=1
                )

        # Create time series plot
        fig, ax = plt.subplots()
        if ftrs_renamed is not None:
            ftrcoef_df.rename(columns=ftrs_renamed).plot(ax=ax, figsize=figsize)
        else:
            ftrcoef_df.plot(ax=ax, figsize=figsize)

        if title is not None:
            plt.title(title)
        else:
            plt.title(f"Feature importances for pipeline: {name}")

        plt.show()

    def intercepts_timeplot(self, name, title=None, figsize=(10, 6)):
        """
        Visualise time series of intercepts for a given pipeline, when available.

        Parameters
        ----------
        name : str
            Name of the previously run signal optimization process.
        title : str, optional
            Title of the plot. Default is None. This creates a figure title of the form
            "Intercepts for pipeline: {name}".
        figsize : tuple of floats or ints, optional
            Tuple of floats or ints denoting the figure size. Default is (10, 6).

        Notes
        -----
        This method displays the time series of intercepts for a given pipeline, when
        available. This information is contained within a line plot.
        """
        # Checks
        if not isinstance(name, str):
            raise TypeError("The pipeline name must be a string.")
        if name not in self.intercepts.name.unique():
            raise ValueError(
                f"""The pipeline name {name} is not in the list of already-calculated 
                pipelines. Please check the pipeline name carefully. If correct, please 
                run calculate_predictions() first.
                """
            )
        intercepts_df = self.get_intercepts(name)

        # TODO: the next line will be made redundament once the signal optimiser checks for this
        # and removes any pipelines with all NaN intercepts
        if intercepts_df.iloc[:, 2:].isna().all().all():
            raise ValueError(
                f"""There are no non-NA intercepts for the pipeline {name}.
                Cannot display a time series plot.
                """
            )
        if not isinstance(title, str) and title is not None:
            raise TypeError("The title must be a string.")
        if not isinstance(figsize, tuple):
            raise TypeError("The figsize argument must be a tuple.")
        if len(figsize) != 2:
            raise ValueError("The figsize argument must be a tuple of length 2.")
        for element in figsize:
            if not isinstance(element, (int, float)):
                raise TypeError(
                    "The elements of the figsize tuple must be floats or ints."
                )

        # Set the style
        sns.set_style("darkgrid")

        # Reshape dataframe for plotting
        intercepts_df = intercepts_df.set_index("real_date")
        intercepts_df = intercepts_df.iloc[:, 1]

        # Create time series plot
        fig, ax = plt.subplots()
        intercepts_df.plot(ax=ax, figsize=figsize)
        if title is not None:
            plt.title(title)
        else:
            plt.title(f"Intercepts for pipeline: {name}")

        plt.show()

    def coefs_stackedbarplot(
        self,
        name,
        ftrs=None,
        title=None,
        cap=None,
        ftrs_renamed=None,
        figsize=(10, 6),
    ):
        """
        Visualise feature coefficients for a given pipeline in a stacked bar plot.

        Parameters
        ----------
        name : str
            Name of the previously run signal optimization process.
        ftrs : list, optional
            List of feature names to plot. Default is None.
        title : str, optional
            Title of the plot. Default is None. This creates a figure title of the form
            "Stacked bar plot of model coefficients: {name}".
        cap : int, optional
            Maximum number of features to display. Default is None. The chosen features
            are the 'cap' most frequently occurring in the pipeline. This cannot exceed
            10.
        ftrs_renamed : dict, optional
            Dictionary to rename the feature names for visualisation in the plot legend.
            Default is None, which uses the original feature names.
        figsize : tuple of floats or ints, optional
            Tuple of floats or ints denoting the figure size. Default is (10, 6).

        Notes
        -----
        This method displays the average feature coefficients for a given pipeline in each
        calendar year, when available. This information is contained within a stacked bar
        plot. The default behaviour is to plot the first 10 features in the order specified
        during training. If more than 10 features were involved in the learning procedure,
        the default is to plot the first 10 features. By specifying a `ftrs` list (which
        can be no longer than 10 elements in length), this default behaviour can be
        overridden.
        """
        # Checks
        if not isinstance(name, str):
            raise TypeError("The pipeline name must be a string.")
        if name not in self.feature_importances.name.unique():
            raise ValueError(
                f"""The pipeline name {name} is not in the list of already-calculated 
                pipelines. Please check the pipeline name carefully. If correct, please 
                run calculate_predictions() first.
                """
            )
        ftrcoef_df = self.get_feature_importances(name)
        if ftrcoef_df.iloc[:, 2:].isna().all().all():
            raise ValueError(
                f"""There are no non-NA coefficients for the pipeline {name}.
                Cannot display a stacked bar plot.
                """
            )
        if ftrs is not None:
            if not isinstance(ftrs, list):
                raise TypeError("The ftrs argument must be a list.")
            if len(ftrs) > 10:
                raise ValueError(
                    "The ftrs list must be no longer than 10 elements in length."
                )
            for ftr in ftrs:
                if not isinstance(ftr, str):
                    raise TypeError("The elements of the ftrs list must be strings.")
                if ftr not in ftrcoef_df.columns:
                    raise ValueError(
                        f"""The feature {ftr} is not in the list of feature coefficients 
                        for the pipeline {name}.
                        """
                    )

        if not isinstance(title, str) and title is not None:
            raise TypeError("The title must be a string.")
        if ftrs_renamed is not None:
            if not isinstance(ftrs_renamed, dict):
                raise TypeError("The ftrs_renamed argument must be a dictionary.")
            for key, value in ftrs_renamed.items():
                if not isinstance(key, str):
                    raise TypeError(
                        "The keys of the ftrs_renamed dictionary must be strings."
                    )
                if not isinstance(value, str):
                    raise TypeError(
                        "The values of the ftrs_renamed dictionary must be strings."
                    )
                if key not in self.X.columns:
                    raise ValueError(
                        f"""The key {key} in the ftrs_renamed dictionary is not a feature 
                        in the pipeline {name}.
                        """
                    )
        if not isinstance(figsize, tuple):
            raise TypeError("The figsize argument must be a tuple.")
        if len(figsize) != 2:
            raise ValueError("The figsize argument must be a tuple of length 2.")
        for element in figsize:
            if not isinstance(element, (int, float, np.int_, np.float_)):
                raise TypeError(
                    "The elements of the figsize tuple must be floats or ints."
                )
        if cap is not None:
            if not isinstance(cap, int):
                raise TypeError("The cap argument must be an integer.")
            if cap <= 0:
                raise ValueError("The cap argument must be greater than zero.")
            if cap > 10:
                raise ValueError("The cap argument must be no greater than 10.")

        # Set the style
        sns.set_style("darkgrid")

        # Reshape dataframe for plotting
        ftrcoef_df = self.get_feature_importances(name)
        years = ftrcoef_df["real_date"].dt.year
        years.name = "year"
        ftrcoef_df.drop(columns=["real_date", "name"], inplace=True)

        # Sort dataframe columns in ascending order of the number of Na values in the columns
        na_count = ftrcoef_df.isna().sum().sort_values()
        reindexed_columns = na_count.index
        ftrcoef_df = ftrcoef_df[reindexed_columns]
        if cap is not None:
            ftrcoef_df = ftrcoef_df.T.head(cap).T
        ftrcoef_df = pd.concat((ftrcoef_df, years), axis=1)

        # Define colour map
        default_cycle_colors = plt.rcParams["axes.prop_cycle"].by_key()["color"][:10]
        cmap = mcolors.LinearSegmentedColormap.from_list(
            "default_cycle", default_cycle_colors
        )

        # Handle case where there are more than 10 features
        if ftrs is not None:
            ftrcoef_df = ftrcoef_df[ftrs + ["year"]]
        else:
            if ftrcoef_df.shape[1] > 11:
                ftrcoef_df = pd.concat(
                    (ftrcoef_df.iloc[:, :10], ftrcoef_df.iloc[:, -1]), axis=1
                )

        # Average the coefficients for each year and separate into positive and negative values
        if ftrs_renamed is not None:
            ftrcoef_df.rename(columns=ftrs_renamed, inplace=True)

        avg_coefs = ftrcoef_df.groupby("year", observed=True).mean()
        pos_coefs = avg_coefs.clip(lower=0)
        neg_coefs = avg_coefs.clip(upper=0)

        ax = None
        # Create stacked bar plot
        if pos_coefs.sum().any():
            ax = pos_coefs.plot(
                kind="bar", stacked=True, figsize=figsize, colormap=cmap, alpha=0.75
            )
        if neg_coefs.sum().any():
            neg_coefs.plot(
                kind="bar",
                stacked=True,
                figsize=figsize,
                colormap=cmap,
                alpha=0.75,
                ax=ax,
            )

        # Display, title, axis labels
        if title is None:
            plt.title(f"Stacked bar plot of model coefficients: {name}")
        else:
            plt.title(title)

        plt.xlabel("Year")
        plt.ylabel("Average Coefficient Value")
        plt.axhline(0, color="black", linewidth=0.8)  # Adds a line at zero

        # Configure legend
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = dict(zip(labels, handles))
        plt.legend(
            by_label.values(),
            by_label.keys(),
            title="Coefficients",
            bbox_to_anchor=(1.05, 1),
            loc="upper left",
        )

        # Display plot
        plt.tight_layout()
        plt.show()

    def nsplits_timeplot(self, name, title=None, figsize=(10, 6)):
        """
        Method to plot the time series for the number of cross-validation splits used
        by the signal optimizer.

        Parameters
        ----------
        name : str
            Name of the previously run signal optimization process.
        title : str, optional
            Title of the plot. Default is None. This creates a figure title of the form
            "Stacked bar plot of model coefficients: {name}".
        figsize : tuple of floats or ints, optional
            Tuple of floats or ints denoting the figure size. Default is (10, 6).
        """
        # Checks
        if not isinstance(name, str):
            raise TypeError("The pipeline name must be a string.")
        if name not in self.chosen_models.name.unique():
            raise ValueError(
                f"""The pipeline name {name} is not in the list of already-calculated 
                pipelines. Please check the pipeline name carefully. If correct, please 
                run calculate_predictions() first.
                """
            )
        models_df = self.get_optimal_models(name)

        if not isinstance(title, str) and title is not None:
            raise TypeError("The title must be a string.")
        if not isinstance(figsize, tuple):
            raise TypeError("The figsize argument must be a tuple.")
        if len(figsize) != 2:
            raise ValueError("The figsize argument must be a tuple of length 2.")
        for element in figsize:
            if not isinstance(element, (int, float)):
                raise TypeError(
                    "The elements of the figsize tuple must be floats or ints."
                )

        # Set the style
        sns.set_style("darkgrid")

        # Reshape dataframe for plotting
        models_df = models_df.set_index("real_date").sort_index()
        models_df = models_df.loc[:, "n_splits_used"]
        models_df_expanded = pd.DataFrame(models_df.tolist(), index=models_df.index)

        # Create time series plot
        # TODO: extend the number of splits line until the first date that the number of splits is incremented
        # This translates into vertical lines at each increment date as opposed to linear interpolation between them.
        fig, ax = plt.subplots()
        models_df_expanded.plot(ax=ax, figsize=figsize)
        if title is not None:
            plt.title(title)
        else:
            plt.title(f"Number of CV splits for pipeline: {name}")

        plt.show()


if __name__ == "__main__":
    from sklearn.linear_model import Ridge, Lasso
    from sklearn.metrics import make_scorer, r2_score, mean_absolute_error
    from macrosynergy.learning import (
        ExpandingKFoldPanelSplit,
        SignWeightedLinearRegression,
        TimeWeightedLinearRegression,
    )
    import scipy.stats as stats
    from macrosynergy.management.simulate import make_qdf
    from macrosynergy.learning.model_evaluation.scorers.scorers import neg_mean_abs_corr
    from macrosynergy.management.types import QuantamentalDataFrame

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2012-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2012-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2012-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2012-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2016, month=1, day=1),
        ),
    }

    so = SignalOptimizer(
        df=dfd,
        xcats=["CRY", "GROWTH", "INFL", "XR"],
        cids=cids,
        blacklist=black,
    )

    so.calculate_predictions(
        name="LR",
        models={
            "Ridge": Ridge(),
            "Lasso": Lasso(),
            "TWLS": TimeWeightedLinearRegression(),
        },
        hyperparameters={
            "Ridge": {
                "fit_intercept": [True, False],
                "alpha": [1e-4, 1e-3, 1e-2, 1e-1, 1, 10, 100, 1000, 10000],
            },
            "Lasso": {
                "fit_intercept": [True, False],
                "alpha": [1e-4, 1e-3, 1e-2, 1e-1, 1, 10, 100, 1000, 10000],
            },
            "TWLS": {
                "half_life": [24, 36, 60, 120, 240],
                "fit_intercept": [True, False],
            },
        },
        scorers={
            "r2": make_scorer(r2_score),
            "mae": make_scorer(mean_absolute_error, greater_is_better=False),
        },
        inner_splitters={
            "ExpandingKFold": ExpandingKFoldPanelSplit(n_splits=5),
            "SecondSplit": ExpandingKFoldPanelSplit(n_splits=10),
        },
        search_type="prior",
        n_iter=6,
        cv_summary="mean-std",
        n_jobs_outer=-1,
        n_jobs_inner=1,
        normalize_fold_results=True,
        split_functions={
            "ExpandingKFold": lambda n: n // 12,
            "SecondSplit": None,
        },
    )

    so.models_heatmap("LR")
    so.feature_importance_timeplot("LR")
    so.coefs_stackedbarplot("LR")
    so.nsplits_timeplot("LR")

    # Test a random forest
    from sklearn.ensemble import RandomForestRegressor

    so.calculate_predictions(
        name="RF",
        models={
            "RF": RandomForestRegressor(
                n_estimators=100,
                min_samples_leaf=5,
                max_features="sqrt",
                max_samples=0.1,
            ),
        },
        hyperparameters={
            "RF": {},
        },
        scorers={
            "r2": make_scorer(r2_score),
            "mae": make_scorer(mean_absolute_error, greater_is_better=False),
        },
        inner_splitters={
            "ExpandingKFold": ExpandingKFoldPanelSplit(n_splits=2),
            "SecondSplit": ExpandingKFoldPanelSplit(n_splits=3),
        },
        search_type="grid",
        cv_summary="mean-std",
        n_jobs_outer=-1,
        n_jobs_inner=1,
    )

    so.feature_importance_timeplot("RF")



import numpy as np
import pandas as pd
import numbers

from sklearn.base import RegressorMixin, BaseEstimator, clone

from macrosynergy.learning.forecasting.bootstrap import BasePanelBootstrap

from abc import ABC


class BaseModifiedRegressor(BaseEstimator, RegressorMixin, BasePanelBootstrap, ABC):
    def __init__(
        self,
        model,
        method,
        error_offset=1e-5,
        bootstrap_method="panel",
        bootstrap_iters=100,
        resample_ratio=1,
        max_features=None,
        analytic_method=None,
    ):
        """
        Modified linear regression model. Estimated coefficients are divided
        by estimated standard errors to form an auxiliary factor model.

        Parameters
        ----------
        model : RegressorMixin
            Underlying linear regression model to be modified to account
            for statistical precision of parameter estimates. This model must
            have `coef_` and `intercept_` attributes, in accordance with
            `scikit-learn` convention.
        method : str
            Method to modify coefficients. Accepted values are
            "analytic" or "bootstrap".
        error_offset : float, default = 1e-5
            Small offset to add to estimated standard errors in order to prevent
            small denominators during the coefficient adjustment.
        bootstrap_method : str, default = "panel"
            Method used to modify coefficients, when `method = bootstrap`.
            Accepted values are "panel", "period", "cross", "cross_per_period",
            "period_per_cross".
        bootstrap_iters : int, default = 100
            Number of bootstrap iterations to determine standard errors, used
            only when `method = bootstrap`.
        resample_ratio : numbers.Number, default = 1
            Ratio of resampling units in each bootstrap dataset, used only
            when `method = bootstrap`. This is a fraction of the quantity of
            the panel component to be resampled.
        max_features : str or int or float, default = None
            Number of features consider in each bootstrap dataset. This is
            used to increase the amount of variation in bootstrap datasets.
            Accepted values are "sqrt", "log2", an integer number of features and
            a floating point proportion of features. Default behaviour is to raise
            a NotImplementedError.
        analytic_method : str, default = None
            The analytic method used to determine standard errors. This parameter
            is passed into `adjust_analyical_se`, which should be implemented
            by the user if analytical, model-specific, expressions are required.

        Notes
        -----
        Parametric regression models are fit by finding optimal parameters that
        minimize a loss function. In the frequentist statistics framework, "true"
        population-wide values exist for these parameters, which can only be
        estimated from sampled data. Consequently, our parameter estimates can be
        considered to be realizations from a random variable, and hence subject to
        sampling variation. Broadly speaking, the greater the amount of independent data
        sampled, the smaller the variation in parameter estimates. In other words,
        parameter estimates are more unreliable when less data is seen during training.
        By estimating the standard deviation of their sampling distributions - a.k.a.
        their "standard errors" - we can adjust our model coefficients to account for
        lack of statistical precision.

        In our modified parametric regression models, each estimated parameter is
        divided by the estimated standard error (plus an offset). This means that greater
        volatility in a parameter estimate due to lack of data is accounted for by
        reducing the magnitude of this estimate, whilst greater certainty in the precision
        of the estimate is reflected by inflating a regression coefficient.

        Use of this class is only recommended for linear models, since these
        regression models are interpretable and the coefficient adjustment can
        accordingly be interpreted as increasing the relevance of factors whose
        coefficients we are more confident in, and decreasing relevance for factors
        whose coefficients we are less confident in. For a more complex function,
        for instance a neural network, amending model coefficients can be disastrous;
        it would be unclear how such adjustment would affect the downstream performance
        of the model. As a consequence, this class should be used with care
        and we recommend its use for linear models only.
        """
        # Checks
        super().__init__(
            bootstrap_method=bootstrap_method,
            resample_ratio=resample_ratio,
            max_features=max_features,
        )

        self._check_init_params(
            model=model,
            method=method,
            error_offset=error_offset,
            bootstrap_iters=bootstrap_iters,
            analytic_method=analytic_method,
        )

        # Set attributes
        self.model = model
        self.method = method
        self.error_offset = error_offset
        self.bootstrap_iters = bootstrap_iters
        self.analytic_method = analytic_method

    def _check_init_params(
        self,
        model,
        method,
        error_offset,
        bootstrap_iters,
        analytic_method,
    ):
        """
        Constructor parameter checks.

        Parameters
        ----------
        model : RegressorMixin
            Underlying linear regression model to be modified to account
            for statistical precision of parameter estimates. This model must
            have `coef_` and `intercept_` attributes, in accordance with
            `scikit-learn` convention.
        method : str
            Method to modify coefficients. Accepted values are
            "analytic" or "bootstrap".
        error_offset : float, default = 1e-5
            Small offset to add to estimated standard errors in order to prevent
            small denominators during the coefficient adjustment.
        bootstrap_iters : int, default = 100
            Number of bootstrap iterations to determine standard errors, used
            only when `method = bootstrap`.
        analytic_method : str, default = None
            The analytic method used to determine standard errors. This parameter
            is passed into `adjust_analyical_se`, which should be implemented
            by the user if analytical, model-specific, expressions are required.
        """
        # model
        if not isinstance(model, BaseEstimator):
            raise TypeError("model must be a valid `scikit-learn` estimator.")
        if not isinstance(model, RegressorMixin):
            raise TypeError("model must be a valid `scikit-learn` regressor.")

        # method
        if not isinstance(method, str):
            raise TypeError("method must be a string.")
        if method not in ["analytic", "bootstrap"]:
            raise ValueError("method must be either 'analytic' or 'bootstrap'.")

        # error_offset
        if not isinstance(error_offset, numbers.Number):
            raise TypeError("error_offset must be a float or an integer.")
        if error_offset <= 0:
            raise ValueError("error_offset must be greater than 0.")

        # bootstrap_iters
        if method == "bootstrap":
            if not isinstance(bootstrap_iters, numbers.Integral):
                raise TypeError("bootstrap_iters must be an integer.")
            if bootstrap_iters <= 0:
                raise ValueError("bootstrap_iters must be a positive integer.")

        # analytic_method
        if method == "analytic":
            if analytic_method is not None:
                if not isinstance(analytic_method, str):
                    raise TypeError("analytic_method must be a string.")

    def fit(
        self,
        X,
        y,
    ):
        """
        Fit a linear model and modify coefficients based on standard errors.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.
        y : pd.DataFrame or pd.Series
            Target vector associated with each sample in X.

        Returns
        -------
        self
            Fitted estimator.
        """
        # Checks
        self._check_fit_params(X=X, y=y)

        # Fit
        self.model.fit(X, y)

        if not hasattr(self.model, "coef_"):
            raise AttributeError("The underlying model must have a `coef_` attribute.")
        if not hasattr(self.model, "intercept_"):
            raise AttributeError(
                "The underlying model must have an `intercept_` attribute."
            )

        # Modify coefficients
        if self.method == "analytic":
            self.intercept_, self.coef_ = self.adjust_analytical_se(
                self.model,
                X,
                y,
                self.analytic_method,
            )
        elif self.method == "bootstrap":
            # clone the model to avoid modifying the original model
            model = clone(self.model)
            self.intercept_, self.coef_ = self.adjust_bootstrap_se(
                model,
                X,
                y,
            )

        return self

    def predict(
        self,
        X,
    ):
        """
        Predict using the unadjusted linear model.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.

        Returns
        -------
        np.ndarray or pd.Series
            Predicted values.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise TypeError("All columns in X must be numeric.")
        if X.isnull().values.any():
            raise ValueError("X must not contain missing values.")
        return self.model.predict(X)

    def create_signal(
        self,
        X,
    ):
        """
        Predict using the coefficient-adjusted linear model.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.

        Returns
        -------
        np.ndarray or pd.Series
            Signal from the adjusted factor model based on X.

        Notes
        -----
        We define an additional `create_signal` method instead of using the
        `predict` method in order to not interfere with hyperparameter
        searches with standard metrics. Moreover, outputs from the adjusted
        factor model are not valid predictions, but are valid trading signals.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise TypeError("All columns in X must be numeric.")
        if X.isnull().values.any():
            raise ValueError("X must not contain missing values.")

        return np.dot(X, self.coef_) + self.intercept_

    def adjust_bootstrap_se(
        self,
        model,
        X,
        y,
    ):
        """
        Adjust the coefficients of the linear model by bootstrap standard errors.

        Parameters
        ----------
        model : RegressorMixin
            The underlying linear model to be modified.
        X : pd.DataFrame
            Input feature matrix.
        y : pd.DataFrame or pd.Series
            Target vector associated with each sample in X.

        Returns
        -------
        intercept : float
            Adjusted intercept.
        coef : np.ndarray
            Adjusted coefficients.
        """
        # Create storage for bootstrap coefficients and intercepts
        bootstrap_coefs = np.zeros((self.bootstrap_iters, X.shape[1]))
        bootstrap_intercepts = np.zeros(self.bootstrap_iters)

        # Bootstrap loop
        for i in range(self.bootstrap_iters):
            X_resampled, y_resampled = self.create_bootstrap_dataset(X, y)
            model.fit(X_resampled, y_resampled)
            bootstrap_coefs[i] = model.coef_
            bootstrap_intercepts[i] = model.intercept_

        # Calculate standard errors
        coef_se = np.std(bootstrap_coefs, axis=0, ddof=0)
        intercept_se = np.std(bootstrap_intercepts, ddof=0)

        # Adjust the coefficients and intercepts by the standard errors
        coef = self.model.coef_ / (coef_se + self.error_offset)
        intercept = self.model.intercept_ / (intercept_se + self.error_offset)

        return intercept, coef

    def adjust_analytical_se(
        self,
        model,
        X,
        y,
        analytic_method,
    ):
        """
        Adjust the coefficients of the linear model by an analytical
        standard error formula.

        Parameters
        ----------
        model : RegressorMixin
            The underlying linear model to be modified.
        X : pd.DataFrame
            Input feature matrix.
        y : pd.DataFrame or pd.Series
            Target vector associated with each sample in X.
        analytic_method : str
            The analytic method used to calculate standard errors.

        Returns
        -------
        intercept : float
            Adjusted intercept.
        coef : np.ndarray
            Adjusted coefficients.

        Notes
        -----
        Analytical standard errors are model-specific, meaning that
        they must be implemented in a subclass of BaseModifiedRegressor.
        """
        raise NotImplementedError(
            "Analytical standard error adjustments are not available for most models."
            "This function must be implemented in a subclass of BaseModifiedRegressor "
            "if known standard error expressions are available."
        )

    def _check_fit_params(
        self,
        X,
        y,
    ):
        """
        Check parameter validity for the fit method.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.
        y : pd.DataFrame or pd.Series
            Target vector associated with each sample in X.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame) and y.shape[1] != 1:
            raise ValueError(
                "The target dataframe must have only one column. If used as part of "
                "an sklearn pipeline, ensure that previous steps return a pandas "
                "series or dataframe."
            )

        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not y.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not y.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't "
                "match."
            )

        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise TypeError("All columns in X must be numeric.")
        if isinstance(y, pd.DataFrame):
            if not pd.api.types.is_numeric_dtype(y.iloc[:, 0]):
                raise TypeError("All columns in y must be numeric.")
        else:
            if not pd.api.types.is_numeric_dtype(y):
                raise TypeError("All columns in y must be numeric.")
        if X.isnull().values.any():
            raise ValueError("X must not contain missing values.")
        if y.isnull().values.any():
            raise ValueError("y must not contain missing values.")



import numpy as np
import pandas as pd

from sklearn.base import BaseEstimator, RegressorMixin

import numbers


class BaseWeightedRegressor(BaseEstimator, RegressorMixin):
    def __init__(
        self,
        model,
        sign_weighted=False,
        time_weighted=False,
        half_life=12 * 21,
    ):
        """
        Weighted regression model to prioritize contribution from certain samples over
        others during training.

        Parameters
        ----------
        model : RegressorMixin
            The underlying model to be trained with weighted sample contributions.
        sign_weighted : bool, optional
            Flag to weight samples based on the frequency of the label's sign in a
            training set.
        time_weighted : bool, optional
            Flag to weight samples based on the recency of the sample.
        half_life : numbers.Number , optional
            Half-life of the exponential decay function used to calculate the time weights.
            This should be expressed in units of the native dataset frequency. Default is
            12 * 21, which corresponds to a half-life of 1-year for a daily dataset.

        Notes
        -----
        Sign-weighted regression models are useful when the dependent (return) variable
        has a directional bias. By assigning higher weights to the less frequent class,
        the model is encouraged to learn equally from both positive and negative return
        samples, irrespective of class imbalance.

        Time-weighted regression models are useful when the practitioner holds the prior
        that more recent samples are more informative than older samples. By assigning
        exponentially decaying weights to samples based on their recency, the model is
        encouraged to prioritize newer information.
        """
        # Checks
        self._check_init_params(model, sign_weighted, time_weighted, half_life)

        # Attributes
        self.model = model
        self.sign_weighted = sign_weighted
        self.time_weighted = time_weighted
        self.half_life = None if not time_weighted else half_life

    def fit(
        self,
        X,
        y,
    ):
        """
        Learn optimal weighted regression parameters.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Pandas dataframe or numpy array of input features.
        y : pd.Series or pd.DataFrame or np.ndarray
            Pandas series, dataframe or numpy array of targets associated with each sample
            in X.
        """
        # Checks
        self._check_fit_params(X, y)

        # Fit
        self.n = X.shape[0]
        self.p = X.shape[1]
        self.sample_weights = self._calculate_sample_weights(y)
        self.model.fit(X, y, sample_weight=self.sample_weights)
        if hasattr(self.model, "coef_"):
            self.coef_ = self.model.coef_
        if hasattr(self.model, "intercept_"):
            self.intercept_ = self.model.intercept_

        return self

    def predict(self, X):
        """
        Predict dependent variable using the fitted weighted regression model.

        Parameters
        ----------
        X : pd.DataFrame or numpy array
            Input feature matrix.

        Returns
        -------
        y_pred : np.ndarray
            Numpy array of predictions.
        """
        # Checks
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Input feature matrix for weighted regressors must be either a pandas "
                "dataframe or numpy array. If used as part of an sklearn pipeline, ensure "
                "that previous steps return a pandas dataframe or numpy array."
            )
        if self.time_weighted and not isinstance(X, pd.DataFrame):
            raise TypeError(
                "When time weighting is enabled, the input feature matrix for weighted "
                "regressors must be a pandas dataframe."
            )
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for weighted regressor forecasts is a "
                    "numpy array, it must have two dimensions. If used as part of an "
                    "sklearn pipeline, ensure that previous steps return a two-dimensional "
                    "data structure."
                )

        if X.shape[1] != self.p:
            raise ValueError(
                "The number of features in the input feature matrix must match the number "
                "seen in training."
            )

        # Predict
        return self.model.predict(X)

    def _calculate_sample_weights(
        self,
        y,
    ):
        """
        Determine sample weights based on the sign and time weighting flags.

        Parameters
        ----------
        y : pd.Series or pd.DataFrame or np.ndarray
            Target vector associated with each sample in X.
        """
        if self.sign_weighted and self.time_weighted:
            sign_weights = self._calculate_sign_weights(y)
            time_weights = self._calculate_time_weights(y)
            return sign_weights * time_weights
        elif self.sign_weighted:
            return self._calculate_sign_weights(y)
        elif self.time_weighted:
            return self._calculate_time_weights(y)
        else:
            return np.ones(y.shape[0])

    def _calculate_sign_weights(
        self,
        targets,
    ):
        """
        Calculate balanced inverse frequency weights for positive and negative signs
        in the target vector.

        Parameters
        ----------
        targets : pd.Series or pd.DataFrame or np.ndarray
            Dependent variable.

        Returns
        -------
        sample_weights : np.ndarray
            Numpy array of sample weights.
        """
        pos_sum = np.sum(targets >= 0)
        neg_sum = np.sum(targets < 0)

        pos_weight = len(targets) / (2 * pos_sum) if pos_sum > 0 else 0
        neg_weight = len(targets) / (2 * neg_sum) if neg_sum > 0 else 0

        sample_weights = np.where(targets >= 0, pos_weight, neg_weight)
        return sample_weights

    def _calculate_time_weights(
        self,
        targets,
    ):
        """
        Calculate exponentially decaying weights based on the recency of the sample in the
        panel.

        Parameters
        ----------
        targets : pd.Series or pd.DataFrame
            Dependent variable.

        Returns
        -------
        sample_weights : np.ndarray
            Numpy array of sample weights.
        """
        dates = sorted(targets.index.get_level_values(1).unique(), reverse=True)
        num_dates = len(dates)
        weights = np.power(2, -np.arange(num_dates) / self.half_life)

        weight_map = dict(zip(dates, weights))
        sample_weights = targets.index.get_level_values(1).map(weight_map).to_numpy()

        return sample_weights

    def _check_init_params(
        self,
        model,
        sign_weighted,
        time_weighted,
        half_life,
    ):
        """
        Checks for constructor parameters.
        """
        if not isinstance(model, RegressorMixin):
            raise TypeError(
                "The model parameter must be an instance of a sklearn regressor."
            )
        if not isinstance(sign_weighted, bool):
            raise TypeError("The sign_weighted parameter must be a boolean.")
        if not isinstance(time_weighted, bool):
            raise TypeError("The time_weighted parameter must be a boolean.")
        if not isinstance(half_life, numbers.Number) or isinstance(half_life, bool):
            raise TypeError("The half_life parameter must be a number.")
        if half_life <= 0:
            raise ValueError("The half_life parameter must be a positive number.")

    def _check_fit_params(
        self,
        X,
        y,
    ):
        """
        Checks for fit method parameters.
        """
        # X
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Input feature matrix for weighted regressors must be either a pandas "
                "dataframe or numpy array."
            )
        elif self.time_weighted and not isinstance(X, pd.DataFrame):
            raise TypeError(
                "When time weighting is enabled, the input feature matrix for weighted "
                "regressors must be a pandas dataframe."
            )
        if isinstance(X, np.ndarray):
            if X.ndim != 2:
                raise ValueError(
                    "When the input feature matrix for weighted regressor forecasts is a "
                    "numpy array, it must have two dimensions."
                )
        # y
        if not isinstance(y, (pd.Series, pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Target vector for weighted regressors must be either a pandas series, "
                "dataframe or numpy array."
            )
        elif self.time_weighted and not isinstance(y, (pd.Series, pd.DataFrame)):
            raise TypeError(
                "When time weighting is enabled, the target vector for weighted regressors "
                "must be either a pandas series or pandas dataframe."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The dependent variable dataframe must have only one column. If used "
                    "as part of an sklearn pipeline, ensure that previous steps return "
                    "a pandas series or dataframe."
                )
        if isinstance(y, np.ndarray):
            if y.ndim != 1:
                raise ValueError(
                    "When the target vector for weighted regressor forecasts is a numpy "
                    "array, it must have one dimension."
                )
        if X.shape[0] != y.shape[0]:
            raise ValueError(
                "The number of samples in the input feature matrix must match the number "
                "of samples in the target vector."
            )

        # Joint X and y checks
        if len(X) != len(y):
            raise ValueError(
                "The number of samples in the input feature matrix must match the number "
                "of samples in the target vector."
            )
        if self.time_weighted:
            if not y.index.equals(X.index):
                raise ValueError(
                    "When time weighting is enabled, the target vector and input feature "
                    "matrix must have the same index."
                )


class SignWeightedRegressor(BaseWeightedRegressor):
    def __init__(self, model):
        """
        Regressor with sign-weighted sample weights.

        Parameters
        ----------
        model : RegressorMixin
            The underlying regression model to be trained with weighted samples.

        Notes
        -----
        By weighting the contribution of different training samples based on the
        sign of the label, the model is encouraged to learn equally from both positive
        and negative return samples, irrespective of class imbalance. If there are more
        positive targets than negative targets in the training set, then the negative
        target samples are given a higher weight in the model training process. The
        opposite is true if there are more negative targets than positive targets.
        """

        super().__init__(model, sign_weighted=True, time_weighted=False)


class TimeWeightedRegressor(BaseWeightedRegressor):
    def __init__(
        self,
        model,
        half_life,
    ):
        """
        Regressor with time-weighted sample weights.

        Parameters
        ----------
        model : RegressorMixin
            The underlying regression model to be trained with weighted samples.
        half_life : numbers.Number
            Half-life of the exponential decay function used to calculate the time weights.

        Notes
        -----
        By weighting the contribution of different training samples based on the
        timestamp, the model is encouraged to prioritise newer information. The half-life
        denotes the number of time periods in units of the native data frequency for the
        weight attributed to the most recent sample to decay by half.
        """
        super().__init__(
            model, sign_weighted=False, time_weighted=True, half_life=half_life
        )



"""
Module for plotting ranges of values across cross-sections for one or more categories.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Tuple, Callable, Optional
from packaging import version

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.types import QuantamentalDataFrame


def view_ranges(
    df: pd.DataFrame,
    xcats: List[str],
    cids: Optional[List[str]] = None,
    start: str = "2000-01-01",
    end: Optional[str] = None,
    val: str = "value",
    kind: str = "bar",
    sort_cids_by: Optional[str] = None,
    title: Optional[str] = None,
    ylab: Optional[str] = None,
    size: Tuple[float] = (16, 8),
    xcat_labels: Optional[List[str]] = None,
    legend_loc: str = "lower center",
    legend_bbox_to_anchor: Optional[Tuple[float]] = None,
):
    """
    Plots averages and various ranges across sections for one or more categories.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame with the necessary columns: 'cid', 'xcat', 'real_date'
        and at least one column with values of interest.
    xcats : List[str]
        extended categories to be checked on. Default is all in the DataFrame.
    cids : List[str]
        cross sections to plot. Default is all in DataFrame.
    start : str
        earliest date in ISO format. Default earliest date in df.
    end : str
        latest date in ISO format. Default is latest date in df.
    val : str
        name of column that contains the values. Default is 'value'.
    kind : str
        type of range plot. Default is 'bar'; other option is 'box'.
    sort_cids_by : str
        criterion for sorting cids on x-axis; Arguments can be 'mean' and 'std'. Default
        is None, i.e. original order. Ordering will be based on the first category if the
        category is defined over the complete panel. Otherwise, mean and standard deviation
        calculated, of the cross-sections, computed across all categories.
    title : str
        string of chart title; defaults depend on type of range plot.
    ylab : str
        y label. Default is no label.
    size : Tuple[float]
        Tuple of width and height of graph. Default is (16, 8).
    xcat_labels : List[str]
        custom labels to be used for the ranges.
    legend_loc : str
        location of legend; passed to matplotlib.pyplot.legend() as `loc`. Default is
        'center right'.
    legend_bbox_to_anchor : Tuple[float]
        passed to matplotlib.pyplot.legend() as `bbox_to_anchor`. Default is None.
        Please see [Matplotlib's Legend Documentation]:
        (https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html) for more
        information on the legend parameters `loc` and `bbox_to_anchor`.
    """

    df = QuantamentalDataFrame(df)

    possible_xcats = set(df["xcat"])
    missing_xcats = set(xcats).difference(possible_xcats)
    error_xcats = (
        "The categories passed in to view_ranges() must be present in the "
        f"DataFrame: missing {missing_xcats}."
    )
    if not set(xcats).issubset(possible_xcats):
        raise ValueError(error_xcats)

    sort_cids_func: Callable = None
    if sort_cids_by is not None:
        if not isinstance(sort_cids_by, str):
            raise TypeError("`sort_cids_by` must be a string.")
        sort_error = "Sorting parameter must either be 'mean' or 'std'."
        if  sort_cids_by not in ["mean", "std"]:
            raise ValueError(sort_error)
        if sort_cids_by == "mean":
            sort_cids_func = np.mean
        else:
            sort_cids_func = np.std

    error_message = (
        "The number of custom labels must match the defined number of "
        "categories in pnl_cats."
    )
    if xcat_labels is not None:
        if len(xcat_labels) != len(xcats):
            raise ValueError(error_message)

    # Unique cross-sections across the union of categories passed - not the intersection.
    # Order of categories will be preserved.
    df, xcats, cids = reduce_df(df, xcats, cids, start, end, out_all=True)

    s_date = df["real_date"].min().strftime("%Y-%m-%d")
    e_date = df["real_date"].max().strftime("%Y-%m-%d")

    sns.set(style="darkgrid")

    if title is None:
        if kind == "bar":
            title = f"Means and standard deviations from {s_date} to {e_date}."
        elif kind == "box":
            title = (
                f"Interquartile ranges, extended ranges and outliers "
                f"from {s_date} to {e_date}."
            )
    if ylab is None:
        ylab = ""

    filt_1 = df["xcat"] == xcats[0]
    first_xcat_cids = set(df[filt_1]["cid"])
    # First category is not defined over all cross-sections.
    order_condition = list(set(cids)) == list(first_xcat_cids)

    if order_condition and sort_cids_func is not None:
        # Sort exclusively on the first category.
        dfx = df[filt_1].groupby(["cid"], observed=True)[val].apply(sort_cids_func)
        order = dfx.sort_values(ascending=False).index
    elif not order_condition and sort_cids_func is not None:
        # Sort across all categories on the available cross-sections.
        dfx = df.groupby(["cid"], observed=True)[val].apply(sort_cids_func)
        order = dfx.sort_values(ascending=False).index
    else:
        order = None

    sns.set_theme(style="darkgrid")
    fig, ax = plt.subplots(figsize=size)

    if kind == "bar":
        if version.parse(sns.__version__) >= version.parse("0.12.0"):
            ax = sns.barplot(
                x="cid",
                y=val,
                hue="xcat",
                hue_order=xcats,
                palette="Paired",
                data=df,
                errorbar="sd",
                order=order,
            )
        else:
            ax = sns.barplot( # pragma: no cover
                x="cid",
                y=val,
                hue="xcat",
                hue_order=xcats,
                palette="Paired",
                data=df,
                ci="sd",
                order=order,
            )
    elif kind == "box":
        ax = sns.boxplot(
            x="cid",
            y=val,
            hue="xcat",
            hue_order=xcats,
            palette="Paired",
            data=df,
            order=order,
        )
        ax.xaxis.grid(True)

    ax.set_title(title, fontdict={"fontsize": 16})
    ax.set_xlabel("")
    ax.set_ylabel(ylab)
    ax.xaxis.grid(True)
    ax.axhline(0, ls="--", linewidth=1, color="black")
    handles, labels = ax.get_legend_handles_labels()

    if xcat_labels is not None:
        labels = xcat_labels

    if (len(xcats) == 1) and (xcat_labels is None):
        ax.get_legend().remove()
    else:
        if legend_bbox_to_anchor is None:
            legend_bbox_to_anchor = (0.5, -0.15 - 0.05 * (len(xcats) - 2))
        ax.legend(
            handles=handles[0:],
            labels=labels[0:],
            loc=legend_loc,
            bbox_to_anchor=legend_bbox_to_anchor,
        )

    plt.tight_layout()

    plt.show()


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD",] = ["2010-01-01", "2020-12-31", 0.5, 0.2]
    df_cids.loc["CAD",] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP",] = ["2012-01-01", "2020-11-30", 0, 2]
    df_cids.loc["USD",] = ["2012-01-01", "2020-11-30", 1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["INFL",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # view_ranges(
    #     dfd,
    #     xcats=["XR"],
    #     kind="box",
    #     start="2012-01-01",
    #     end="2018-01-01",
    #     sort_cids_by="std",
    # )

    filter_1 = (dfd["xcat"] == "XR") & (dfd["cid"] == "AUD")
    dfd = dfd[~filter_1]

    view_ranges(
        dfd,
        xcats=["XR", "CRY"],
        cids=cids,
        kind="box",
        start="2012-01-01",
        end="2018-01-01",
        sort_cids_by=None,
        xcat_labels=["EQXR_NSA", "CRY_NSA"],
    )



"""
Module hosting custom types and meta-classes for use across the package.
"""

from typing import List
import pandas as pd


class QuantamentalDataFrameMeta(type):
    """
    Metaclass to support type checks for `QuantamentalDataFrame`.
    """

    IndexCols: List[str] = ["real_date", "cid", "xcat"]
    _StrIndexCols: List[str] = ["cid", "xcat"]

    def __instancecheck__(cls, instance):
        IDX_COLS = QuantamentalDataFrameMeta.IndexCols
        try:
            # the try except offers a safety net in case the instance is not a
            # pd.DataFrame and one of the checks raises an error
            result = isinstance(instance, pd.DataFrame)
            result = result and instance.index.name is None
            result = result and not isinstance(instance.columns, pd.MultiIndex)
            result = result and all([col in instance.columns for col in IDX_COLS])
            result = result and len(instance.columns) > len(IDX_COLS)
            result = result and len(instance.columns) == len(set(instance.columns))

            datetypestr = str(instance["real_date"].dtype)
            is_valid_dt64 = (datetypestr in ["datetime64[ns]", "datetime64[ms]"]) or (
                datetypestr.startswith("datetime64[") and datetypestr.endswith("]")
            )
            correct_date_type: bool = (
                is_valid_dt64
                or isinstance(instance["real_date"].dtype, pd.DatetimeTZDtype)
                or instance.empty
            )
            result = result and correct_date_type

        except:  # noqa
            result: bool = False
        finally:
            return result


class QuantamentalDataFrameBase(pd.DataFrame, metaclass=QuantamentalDataFrameMeta):
    """
    Base class to extend `pd.DataFrame` for Quantamental DataFrames.  This class is a
    parent class to macrosynergy.types.QuantamentalDataFrame.
    """

    IndexCols: List[str] = ["real_date", "cid", "xcat"]
    _StrIndexCols: List[str] = ["cid", "xcat"]
    IndexColsSortOrder = ["cid", "xcat", "real_date"]
